/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#ifndef _ArtTextP_h
#define _ArtTextP_h

#include "ArtText.h"
#include "ShadowP.h"

typedef struct {
    XtPointer	extension;
} ArtTextClassPart;

typedef struct _ArtTextClassRec {
    CoreClassPart       core_class;
    ShadowClassPart	shadow_class;
    ArtTextClassPart	arttext_class;
} ArtTextClassRec;

extern ArtTextClassRec artTextClassRec;

struct line_table_item {
    LineType		type;
    long		y;
    union {
	struct {
	    char	*string;
	    long	no_offsets;
	    long	*offsets;
	} str;
	struct {
	    Pixmap	pixmap;
	    Dimension	width;
	    Dimension	height;
	} img;
	struct {
	    Dimension	size;
	} sep;
	struct {
	    char		*string;
	    XtCallbackProc	callback;
	    XtPointer		client_data;
	} cli;
	struct {
	    Dimension		size;
	} spc;
    } u;
};

typedef struct {
    Pixel               foreground_pixel;
    Pixel		quote_pixel;
    Pixel		header_pixel;
    Pixel		clickable_pixel;
    Pixel		highlight_fg;
    Pixel		highlight_bg;
    XFontStruct         *font;
    XFontStruct		*quote_font;
    XFontStruct		*header_font;
    XFontStruct		*clickable_font;
    long		first;
    long		shown;
    long		lines;
    long		n_alloc;
    XtCallbackList	resize_callback;
    Dimension		image_margin;
    Dimension		separator_margin;
    Dimension		preferred_lines;
    Dimension		preferred_columns;
    /* private data */
    struct line_table_item	*line_table;
    GC                  ordinary_gc;
    GC			ordinary_h_gc;
    GC			quote_gc;
    GC			quote_h_gc;
    GC			header_gc;
    GC			header_h_gc;
    GC			clickable_gc;
    GC			highlight_gc;
    Atom		curr_sel;
    Time		sel_time;
    long		sel_start_line;
    long		sel_stop_line;
    int			sel_start_offset;
    int			sel_stop_offset;
    Boolean		sel_ok;
    Boolean		extending;
    Boolean		extend_end;
} ArtTextPart;

typedef struct _ArtTextRec {
    CorePart    core;
    ShadowPart	shadow;
    ArtTextPart	arttext;
} ArtTextRec;

#endif /* KJJ_ArtTextP_h */
