/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */

/*

Copyright (c) 1989  X Consortium

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall
not be used in advertising or otherwise to promote the sale, use or
other dealings in this Software without prior written authorization
from the X Consortium.

*/

#define NEED_EVENTS
#define NEED_REPLIES
#include <X11/Xproto.h>
#include <X11/Xlibint.h>
#include <X11/IntrinsicP.h>

#include "Compat.h"

/* This function is modified from the source to XTextWidth,
 * which is why I included the X Consortium copyright notice.
 */
int MyXWidthToChars(XFontStruct *font, char *string, int len, int width)
{
    unsigned char	*ustr = (unsigned char *)string;
    XCharStruct		*def;
    Bool		singlerow = (font->max_byte1 == 0);
    int			n = len;

    if (!ustr || width <= 0)
	return 0;

    if (singlerow) {
	CI_GET_DEFAULT_INFO_1D(font, def);
    } else {
	CI_GET_DEFAULT_INFO_2D(font, def);
    }

    while (n > 0) {
	XCharStruct	*cs;
	unsigned int	c = (unsigned int)*ustr;

	if (singlerow) {
	    CI_GET_CHAR_INFO_1D(font, c, def, cs);
	} else {
	    /* essentially the macro        */
	    /* CI_GET_ROWZERO_CHAR_INFO_2D  */
	    if (font->min_byte1 == 0 &&
		c >= font->min_char_or_byte2 &&
		c <= font->max_char_or_byte2) {
		if (font->per_char) {
		    cs = &font->per_char[(c - font->min_char_or_byte2)];
		    if (CI_NONEXISTCHAR(cs))
			cs = def;
		} else {
		    cs = &font->min_bounds;
		}
	    } else {
		cs = def;
	    }
	}

	if (cs)
	    width -= cs->width;
	if (width < 0)
	    break; /* not <= ... */
	n--;
	ustr++;
    }	

    return len - n;
}

/*************************************************************************/

struct callback_entry {
    XtCallbackProc	proc;
    XtPointer		client_data;
    Atom		wm_delete_window;
    Atom		wm_protocols;
};

static void WM_DELETE_WINDOW_handler(Widget w, XtPointer client_data,
				     XEvent *event, Boolean *cont)
{
    struct callback_entry	*entry = (struct callback_entry *)client_data;

    if (event->type == ClientMessage &&
	event->xclient.message_type == entry->wm_protocols &&
	event->xclient.data.l[0] == entry->wm_delete_window) {
	entry->proc(w, entry->client_data, (XtPointer)event);
	*cont = False;
    }
}

void add_WM_DELETE_WINDOW_callback(Widget w,
				   XtCallbackProc proc,
				   XtPointer client_data)
{
    Display	*disp = XtDisplay(w);
    Window	win = XtWindow(w);
    Atom	wm_delete_window, wm_protocols;
    struct callback_entry	*entry;

    wm_protocols = XInternAtom(disp, "WM_PROTOCOLS", False);
    wm_delete_window = XInternAtom(disp, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(disp, win, &wm_delete_window, 1);

    entry = (struct callback_entry *)XtMalloc(sizeof(struct callback_entry));
    entry->proc = proc;
    entry->client_data = client_data;
    entry->wm_protocols = wm_protocols;
    entry->wm_delete_window = wm_delete_window;

    XtAddEventHandler(w, (EventMask)0, True,
		      WM_DELETE_WINDOW_handler, (XtPointer)entry);
}

/*************************************************************************/

int get_event_xy(XEvent *event, int *x, int *y)
{
    if (!event)
	return False;

    switch (event->type) {
    case ButtonPress:
    case ButtonRelease:
        *x = event->xbutton.x;
        *y = event->xbutton.y;
        return True;
    case KeyPress:
    case KeyRelease:
        *x = event->xkey.x;
        *y = event->xkey.y;
        return True;
    case EnterNotify:
    case LeaveNotify:
        *x = event->xcrossing.x;
        *y = event->xcrossing.y;
        return True;
    case MotionNotify:
        *x = event->xmotion.x;
        *y = event->xmotion.y;
        return True;
    }

    return False;
}

Time get_event_time(XEvent *event)
{
    switch (event->type) {
    case ButtonPress:
    case ButtonRelease:
	return event->xbutton.time;
    case KeyPress:
    case KeyRelease:
	return event->xkey.time;
    case EnterNotify:
    case LeaveNotify:
	return event->xcrossing.time;
    case MotionNotify:
	return event->xmotion.time;
    case PropertyNotify:
	return event->xproperty.time;
    case SelectionClear:
	return event->xselectionclear.time;
    case SelectionNotify:
	return event->xselection.time;
    case SelectionRequest:
	return event->xselectionrequest.time;
    }

    return CurrentTime;
}
