/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <pwd.h>

#include "../Widgets/Compat.h"
#include "GroupObjeP.h"

#define XtRHeaderFormat	"HeaderFormat"
#define XtRKeepType	"KeepType"

static char default_header_format[] =
"Subject:Newsgroups:Followup-To:Reply-To:"
"Content-Description:Date:Organization:From:";
static char default_attribution[] =
"In article %m,\n\t%f writes:";
static char default_signature[] =
".signature";

static XtResource resources[] = {
#define offset(field) XtOffsetOf(GroupObjectRec, group_object.field)
    {XtNheaderFormat, XtCHeaderFormat, XtRHeaderFormat, sizeof(String*),
     offset(header_format), XtRString, (XtPointer)default_header_format},
    {XtNquoteRegexp, XtCQuoteRegexp, XtRString, sizeof(String),
     offset(quote_regexp), XtRImmediate, (XtPointer)"^[:>]"},
    {XtNquoteString, XtCQuoteString, XtRString, sizeof(String),
     offset(quote_string), XtRImmediate, (XtPointer)">"},
    {XtNattribution, XtCAttribution, XtRString, sizeof(String),
     offset(attribution), XtRImmediate, (XtPointer)default_attribution},
    {XtNfullName, XtCFullName, XtRString, sizeof(String),
     offset(full_name), XtRImmediate, (XtPointer)NULL},
    {XtNsignatureFile, XtCSignatureFile, XtRString, sizeof(String),
     offset(signature_file), XtRImmediate, (XtPointer)default_signature},
    {XtNorganization, XtCOrganization, XtRString, sizeof(String),
     offset(organization), XtRImmediate, (XtPointer)NULL},
    {XtNreplyTo, XtCReplyTo, XtRString, sizeof(String),
     offset(reply_to), XtRImmediate, (XtPointer)NULL},
    {XtNextraHeaders, XtCExtraHeaders, XtRString, sizeof(String),
     offset(extra_headers), XtRImmediate, (XtPointer)NULL},
    {XtNuuDir, XtCUuDir, XtRString, sizeof(String),
     offset(uu_dir), XtRImmediate, (XtPointer)NULL},
    {XtNuuProgram, XtCUuProgram, XtRString, sizeof(String),
     offset(uu_program), XtRImmediate, (XtPointer)NULL},
    {XtNdefaultHotColor, XtCForeground, XtRPixel, sizeof(Pixel),
     offset(default_hot_pixel), XtRString, XtDefaultForeground},
    {XtNfullHeader, XtCFullHeader, XtRBoolean, sizeof(Boolean),
     offset(full_header), XtRImmediate, (XtPointer)False},
    {XtNprocessXrefs, XtCProcessXrefs, XtRBoolean, sizeof(Boolean),
     offset(process_xrefs), XtRImmediate, (XtPointer)True},
    {XtNshowNumberLines, XtCShowNumberLines, XtRBoolean, sizeof(Boolean),
     offset(show_number_lines), XtRImmediate, (XtPointer)False},
    {XtNonlyUnread, XtCOnlyUnread, XtRBoolean, sizeof(Boolean),
     offset(only_unread), XtRImmediate, (XtPointer)False},
    {XtNkeepThreadInfo, XtCKeepThreadInfo, XtRKeepType, sizeof(KeepType),
     offset(keep_thread_info), XtRImmediate, (XtPointer)KeepTypeNo},
    {XtNfixedAuthor, XtCFixedAuthor, XtRBoolean, sizeof(Boolean),
     offset(fixed_author), XtRImmediate, (XtPointer)False},
#undef offset
};

static void ClassInitialize(void);
static void Initialize(Widget, Widget, ArgList, Cardinal*);
static void Destroy(Widget);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);

GroupObjectClassRec groupObjectClassRec = {
    {				/* object fields	*/
	(WidgetClass)&objectClassRec, /* superclass	*/
	"GroupObject",		/* class_name		*/
	sizeof(GroupObjectRec),	/* widget_size		*/
	ClassInitialize,	/* class_initialize	*/
	NULL,			/* class_part_initialize*/
	FALSE,			/* class_inited		*/
	Initialize,		/* initialize		*/
	NULL,			/* initialize_hook	*/
	NULL,			/* pad			*/
	NULL,			/* pad			*/
	0,			/* pad			*/
	resources,		/* resources		*/
	XtNumber(resources),	/* num_resources	*/
	NULLQUARK,		/* xrm_class		*/
	FALSE,			/* pad			*/
	FALSE,			/* pad			*/
	FALSE,			/* pad			*/
	FALSE,			/* pad			*/
	Destroy,		/* destroy		*/
	NULL,			/* pad			*/
	NULL,			/* pad			*/
	SetValues,		/* set_values		*/
	NULL,			/* set_values_hook	*/
	NULL,			/* pad			*/
	NULL,			/* get_values_hook	*/
	NULL,			/* pad			*/
	XtVersion,		/* version		*/
	NULL,			/* callback_offsets	*/
	NULL,			/* pad			*/
	NULL,			/* pad			*/
	NULL,			/* pad			*/
	NULL,			/* extension		*/
    },
    {				/* group_object fields	*/
	NULL,			/* empty		*/
    }
};

WidgetClass groupObjectClass = (WidgetClass)&groupObjectClassRec;

/*************************************************************************/

static Boolean CvtStringToHeaderFormat(Display *disp, XrmValue *args,
				       Cardinal *no_args, XrmValue *from,
				       XrmValue *to, XtPointer *conv_data)
{
    static String	*result;
    String		string;
    char		*c;
    int			n;

    if (to->addr)
	if (to->size < sizeof(String*)) {
	    to->size = sizeof(String*);
	    return False;
	}

    string = (String)from->addr;
    for (c = string, n = 0 ; c ; c = strchr(c + 1, ':'), n++);
    result = (String *)XtMalloc((n + 2) * sizeof(String));

    for (n = 0, c = string ; *c != '\0' ; n++) {
	char	*p = strchr(c, ':');
	int	len;

	if (!p)
	    break;

	len = p - c + 1;
	result[n] = XtMalloc(len + 1);
	memcpy(result[n], c, len);
	result[n][len] = '\0';
	c = p + 1;
    }
    result[n] = NULL;

    if (to->addr)
	*(String **)(to->addr) = result;
    else
	to->addr = (XPointer)&result;
    to->size = sizeof(String*);

    return True;
}

static void FreeHeaderFormat(XtAppContext app_cont, XrmValue *to,
			     XtPointer conv_data,
			     XrmValue *args, Cardinal *no_args)
{
    if (to->addr) {
	String	*format, *loop;

	format = loop = (String *)to->addr;
	while (*loop) {
	    XtFree(*loop);
	    loop++;
	}
	XtFree((char *)format);
    }
}

static Boolean CvtStringToKeepType(Display *disp, XrmValue *args,
				   Cardinal *no_args, XrmValue *from,
				   XrmValue *to, XtPointer *conv_data)
{
    static XrmQuark	yes   = NULLQUARK;
    static XrmQuark	true  = NULLQUARK;
    static XrmQuark	no    = NULLQUARK;
    static XrmQuark	false = NULLQUARK;
    static XrmQuark	subs  = NULLQUARK;
    static KeepType	result;
    char		*str = (char *)from->addr;
    char		buffer[32];
    char		*c;
    XrmQuark		q;

    if (to->addr && to->size < sizeof(KeepType)) {
	to->size = sizeof(KeepType);
	return False;
    }

    if (yes == NULLQUARK) {
	yes   = XrmPermStringToQuark("yes");
	true  = XrmPermStringToQuark("true");
	no    = XrmPermStringToQuark("no");
	false = XrmPermStringToQuark("false");
	subs  = XrmPermStringToQuark("subscribed");
    }

    if (!str || strlen(str) >= 32)
	return False;

    c = buffer;
    while (*str != '\0')
	if (*str < 'A' || *str > 'Z')
	    *c++ = *str++;
	else
	    *c++ = *str++ - 'A' + 'a';
    *c = '\0';

    q = XrmStringToQuark(buffer);

    if (q == yes || q == true)
	result = KeepTypeYes;
    else if (q == no || q == false)
	result = KeepTypeNo;
    else if (q == subs)
	result = KeepTypeSubscribed;
    else
	return False;

    if (to->addr)
	*(KeepType *)to->addr = result;
    else
	to->addr = (XPointer)&result;
    to->size = sizeof(KeepType);

    return True;
}

/*************************************************************************/

static void ClassInitialize()
{
    XtSetTypeConverter(XtRString, XtRHeaderFormat, CvtStringToHeaderFormat,
		       NULL, 0, XtCacheAll, FreeHeaderFormat);
    XtSetTypeConverter(XtRString, XtRKeepType, CvtStringToKeepType,
		       NULL, 0, XtCacheNone, NULL);
}

static void Initialize(Widget grequest, Widget gnew,
		       ArgList args, Cardinal *num_args)
{
    GroupObject	new = (GroupObject)gnew;
    char	buffer[1024];
    char	*home = getenv("HOME");

    if (new->group_object.signature_file) {
	if (new->group_object.signature_file[0] == '\0')
	    new->group_object.signature_file = NULL;
	else if (!home || new->group_object.signature_file[0] == '/')
	    new->group_object.signature_file =
		XtNewString(new->group_object.signature_file);
	else {
	    sprintf(buffer, "%s/%s", home, new->group_object.signature_file);
	    new->group_object.signature_file = XtNewString(buffer);
	}
    }

    new->group_object.comp_quote_re = NULL;
    if (new->group_object.quote_regexp) {
	regex_t	re;

	new->group_object.quote_regexp =
	    XtNewString(new->group_object.quote_regexp);
	if (regcomp(&re, new->group_object.quote_regexp,
		    REG_EXTENDED|REG_NOSUB) == 0) {
	    new->group_object.comp_quote_re =
		(regex_t *)XtMalloc(sizeof(regex_t));
	    (void)memcpy(new->group_object.comp_quote_re,
			 &re, sizeof(regex_t));
	} else {
	    sprintf(buffer, "GroupObject: Failed to compile regexp: %s\n",
		    new->group_object.quote_regexp);
	    XtAppWarning(XtWidgetToApplicationContext((Widget)new), buffer);
	    regfree(&re);
	}
    }

    if (!new->group_object.organization) {
	char	*env_var;

	env_var = getenv("NEWSORG");
	if (!env_var) env_var = getenv("ORGANIZATION");

	if (env_var) new->group_object.organization = XtNewString(env_var);
    } else if (new->group_object.organization[0] != '\0')
	new->group_object.organization =
	    XtNewString(new->group_object.organization);
    else
	new->group_object.organization = NULL;

    if (new->group_object.uu_dir) {
	if (new->group_object.uu_dir[0] == '/' || !home)
	    new->group_object.uu_dir = XtNewString(new->group_object.uu_dir);
	else {
	    sprintf(buffer, "%s/%s", home, new->group_object.uu_dir);
	    new->group_object.uu_dir = XtNewString(buffer);
	}
    }
}

static void Destroy(Widget gw)
{
    GroupObject	w = (GroupObject)gw;

    XtFree(w->group_object.quote_regexp);
    if (w->group_object.comp_quote_re) {
	regfree(w->group_object.comp_quote_re);
	XtFree((char *)w->group_object.comp_quote_re);
    }
    XtFree(w->group_object.signature_file);
    XtFree(w->group_object.organization);
    XtFree(w->group_object.uu_dir);
}

static Boolean SetValues(Widget gcurrent,
			 Widget grequest,
			 Widget gnew,
			 ArgList args,
			 Cardinal *num_args)
{
    GroupObject	new = (GroupObject)gnew;
    GroupObject	current = (GroupObject)gcurrent;

    if (new->group_object.quote_regexp !=
	current->group_object.quote_regexp) {
	XtFree(current->group_object.quote_regexp);
	if (current->group_object.comp_quote_re) {
	    regfree(current->group_object.comp_quote_re);
	    XtFree((char *)current->group_object.comp_quote_re);
	    current->group_object.comp_quote_re = NULL;
	}
	
	if (new->group_object.quote_regexp) {
	    regex_t	re;

	    new->group_object.quote_regexp =
		XtNewString(new->group_object.quote_regexp);
	    if (regcomp(&re, new->group_object.quote_regexp,
			REG_NOSUB|REG_EXTENDED) == 0) {
		new->group_object.comp_quote_re =
		    (regex_t *)XtMalloc(sizeof(regex_t));
		(void)memcpy(new->group_object.comp_quote_re,
			     &re, sizeof(regex_t));
	    } else {
		char	buffer[1024];

		sprintf(buffer, "GroupObject: Failed to compile regexp: %s\n",
			new->group_object.quote_regexp);
		XtAppWarning(XtWidgetToApplicationContext((Widget)new),
			     buffer);
		regfree(&re);
	    }
	}
    }

    return False;
}

/*************************************************************************/

char **GroupObjectHeaderFormat(GroupObject w)
{
    if (w)
	return w->group_object.header_format;
    else
	return NULL;
}

regex_t *GroupObjectQuoteRegexp(GroupObject w)
{
    if (w)
	return w->group_object.comp_quote_re;
    else
	return NULL;
}

char *GroupObjectQuoteString(GroupObject w)
{
    if (w)
	return w->group_object.quote_string;
    else
	return NULL;
}

char *GroupObjectAttribution(GroupObject w)
{
    if (w)
	return w->group_object.attribution;
    else
	return NULL;
}

char *GroupObjectFullName(GroupObject w)
{
    if (w)
	return w->group_object.full_name;
    else
	return NULL;
}

char *GroupObjectSignatureFile(GroupObject w)
{
    if (w)
	return w->group_object.signature_file;
    else
	return NULL;
}

char *GroupObjectOrganization(GroupObject w)
{
    if (w)
	return w->group_object.organization;
    else
	return NULL;
}

char *GroupObjectReplyTo(GroupObject w)
{
    if (w)
	return w->group_object.reply_to;
    else
	return NULL;
}

char *GroupObjectExtraHeaders(GroupObject w)
{
    if (w)
	return w->group_object.extra_headers;
    else
	return NULL;
}

char *GroupObjectUuDir(GroupObject w)
{
    if (w)
	return w->group_object.uu_dir;
    else
	return NULL;
}

char *GroupObjectUuProgram(GroupObject w)
{
    if (w)
	return w->group_object.uu_program;
    else
	return NULL;
}

Pixel GroupObjectDefaultHotPixel(GroupObject w)
{
    if (w)
	return w->group_object.default_hot_pixel;
    else
	return 0;
}

Boolean GroupObjectFullHeader(GroupObject w)
{
    if (w)
	return w->group_object.full_header;
    else
	return False;
}

Boolean GroupObjectProcessXrefs(GroupObject w)
{
    if (w)
	return w->group_object.process_xrefs;
    else
	return False;
}

Boolean GroupObjectShowNumberLines(GroupObject w)
{
    if (w)
	return w->group_object.show_number_lines;
    else
	return False;
}

Boolean GroupObjectOnlyUnread(GroupObject w)
{
    if (w)
	return w->group_object.only_unread;
    else
	return False;
}

KeepType GroupObjectKeepThreadInfo(GroupObject w)
{
    if (w)
	return w->group_object.keep_thread_info;
    else
	return False;
}

Boolean GroupObjectFixedAuthor(GroupObject w)
{
    if (w)
	return w->group_object.fixed_author;
    else
	return False;
}

GroupObject GroupObjectCreate(Widget parent, char *newsgroup)
{
    char	name[1024];
    char	*c;

    sprintf(name, "%s", newsgroup);
    for (c = name ; *c != '\0' ; c++)
	if (*c == '.' || *c == ':')
	    *c = '_';

    /* XtNewString(name) ... */
    return (GroupObject)XtCreateWidget(name, groupObjectClass,
				       parent, NULL, 0);
}
