/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/CharSet.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "../Widgets/Compat.h"
#include "NntpObjecP.h"

#define XtRGroupList	"GroupList"
#define XtRReconnect	"Reconnect"

static char default_auto_subscribe[] =
"news.answers news.announce.newusers";

static XtResource resources[] = {
#define offset(field) XtOffsetOf(NntpObjectRec, nntp_object.field)
    {XtNnntpServer, XtCNntpServer, XtRString, sizeof(String),
     offset(nntp_server), XtRImmediate, (XtPointer)NULL},
    {XtNnewsrcDir, XtCNewsrcDir, XtRString, sizeof(String),
     offset(newsrc_dir), XtRImmediate, (XtPointer)""},
    {XtNnewsrcFile, XtCNewsrcFile, XtRString, sizeof(String),
     offset(newsrc_file), XtRImmediate, (XtPointer)NULL},
    {XtNoldNewsrcFile, XtCOldNewsrcFile, XtRString, sizeof(String),
     offset(old_newsrc_file), XtRImmediate, (XtPointer)NULL},
    {XtNkillDir, XtCKillDir, XtRString, sizeof(String),
     offset(kill_dir), XtRImmediate, (XtPointer)""},
    {XtNkillFile, XtCKillFile, XtRString, sizeof(String),
     offset(kill_file), XtRImmediate, (XtPointer)NULL},
    {XtNautoSubscribe, XtCAutoSubscribe, XtRGroupList, sizeof(String*),
     offset(auto_subscribe), XtRString, (XtPointer)default_auto_subscribe},
    {XtNauthInfoUser, XtCAuthInfoUser, XtRString, sizeof(String),
     offset(auth_info_user), XtRImmediate, (XtPointer)NULL},
    {XtNauthInfoPass, XtCAuthInfoPass, XtRString, sizeof(String),
     offset(auth_info_pass), XtRImmediate, (XtPointer)NULL},
    {XtNauthInfoGeneric, XtCAuthInfoGeneric, XtRString, sizeof(String),
     offset(auth_info_generic), XtRImmediate, (XtPointer)NULL},
    {XtNpostingAgent, XtCPostingAgent, XtRString, sizeof(String),
     offset(posting_agent), XtRImmediate, (XtPointer)NULL},
    {XtNrescanTimeout, XtCRescanTimeout, XtRInt, sizeof(int),
     offset(rescan_timeout), XtRImmediate, (XtPointer)60},
    {XtNcreateNewsrc, XtCCreateNewsrc, XtRBoolean, sizeof(Boolean),
     offset(create_newsrc), XtRImmediate, (XtPointer)True},
    {XtNretrieveDescriptions, XtCRetrieveDescriptions,
     XtRBoolean, sizeof(Boolean),
     offset(retrieve_descriptions), XtRImmediate, (XtPointer)True},
    {XtNcheckForNewGroups, XtCCheckForNewGroups, XtRBoolean, sizeof(Boolean),
     offset(check_for_new_groups), XtRImmediate, (XtPointer)True},
    {XtNreadActiveFile, XtCReadActiveFile, XtRBoolean, sizeof(Boolean),
     offset(read_active_file), XtRImmediate, (XtPointer)True},
    {XtNfillNewsrcFile, XtCFillNewsrcFile, XtRBoolean, sizeof(Boolean),
     offset(fill_newsrc_file), XtRImmediate, (XtPointer)False},
    {XtNthreadEmAll, XtCThreadEmAll, XtRBoolean, sizeof(Boolean),
     offset(thread_em_all), XtRImmediate, (XtPointer)False},
#undef offset
};

static void ClassInitialize(void);
static void Initialize(Widget, Widget, ArgList, Cardinal*);
static void Destroy(Widget);
static void Realize(Widget, XtValueMask*, XSetWindowAttributes*);
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal*);

static CompositeClassExtensionRec	compositeClassExt = {
    NULL,				/* next_extension		*/
    NULLQUARK,				/* record_type			*/
    XtCompositeExtensionVersion,	/* version			*/
    sizeof(CompositeClassExtensionRec),	/* record_size			*/
    TRUE,				/* accepts_objects		*/
#if (XT_REVISION >= 6)
    FALSE,				/* allows_change_managed_set	*/
#endif
};

NntpObjectClassRec nntpObjectClassRec = {
    {					/* core fields		*/
	(WidgetClass)&compositeClassRec,/* superclass		*/
	"NntpObject",			/* class_name		*/
	sizeof(NntpObjectRec),		/* widget_size		*/
	ClassInitialize,		/* class_initialize	*/
	NULL,				/* class_part_initialize*/
	FALSE,				/* class_inited		*/
	Initialize,			/* initialize		*/
	NULL,				/* initialize_hook	*/
	Realize,			/* realize		*/
	NULL,				/* actions		*/
	0,				/* num_actions		*/
	resources,			/* resources		*/
	XtNumber(resources),		/* num_resources	*/
	NULLQUARK,			/* xrm_class		*/
	FALSE,				/* compress_motion	*/
	FALSE,				/* compress_exposure	*/
	FALSE,				/* compress_enterleave	*/
	FALSE,				/* visible_interest	*/
	Destroy,			/* destroy		*/
	NULL,				/* resize		*/
	NULL,				/* expose		*/
	SetValues,			/* set_values		*/
	NULL,				/* set_values_hook	*/
	NULL,				/* set_values_almost	*/
	NULL,				/* get_values_hook	*/
	NULL,				/* accept_focus		*/
	XtVersion,			/* version		*/
	NULL,				/* callback_offsets	*/
	NULL,				/* tm_table		*/
	NULL,				/* query_geometry	*/
	NULL,				/* display_accelerator	*/
	NULL,				/* extension		*/
    },
    {					/* composite fields	*/
	XtInheritGeometryManager,	/* geometry_manager	*/
	XtInheritChangeManaged,		/* change_managed	*/
	XtInheritInsertChild,		/* insert_child		*/
	XtInheritDeleteChild,		/* delete_child		*/
	(XtPointer)&compositeClassExt,	/* extension		*/
    },
    {					/* newsobj fields	*/
	NULL,				/* empty		*/
    }
};

WidgetClass nntpObjectClass = (WidgetClass)&nntpObjectClassRec;

/*************************************************************************/

static Boolean CvtStringToGroupList(Display *disp, XrmValue *args,
				    Cardinal *num_args, XrmValue *from,
				    XrmValue *to, XtPointer *client_data)
{
    static String	*result;
    String		temp;
    int			n;
    char		*c;

    if (to->addr && to->size < sizeof(String*)) {
	to->size = sizeof(String*);
	return False;
    } else if (!from->addr) {
	if (to->addr)
	    *(String*)to->addr = NULL;
	else
	    to->addr = NULL;
	to->size = sizeof(String*);
	return True;
    }

    c = (char *)from->addr;
    while (isspace((unsigned char)*c))
	c++;
    temp = XtNewString(c);
    for (n = 0 ; *c != '\0' ; n++) {
	while (*c != '\0' && !isspace((unsigned char)*c))
	    c++;
	while (isspace((unsigned char)*c))
	    c++;
    }

    if (n == 0) {
	XtFree(temp);
	*(String*)to->addr = NULL;
	return True;
    }

    result = (String *)XtMalloc((n + 1) * sizeof(String));
    for (c = temp, n = 0 ; *c != '\0'  ; n++) {
	result[n] = c;
	while (*c != '\0' && !isspace((unsigned char)*c))
	    c++;
	if (*c != 0) *c++ = '\0';

	while (isspace((unsigned char)*c))
	    c++;
    }
    result[n] = NULL;

    if (to->addr)
	*(String**)to->addr = result;
    else
	to->addr = (XPointer)&result;
    to->size = sizeof(String*);

    return True;
}

static void FreeGroupList(XtAppContext app_cont, XrmValue *to,
			  XtPointer data, XrmValue *args, Cardinal *num_args)
{
    if (to->addr)
	XtFree(((char **)to->addr)[0]);
    XtFree((char *)to->addr);
}

/*************************************************************************/

static void ClassInitialize()
{
    XtSetTypeConverter(XtRString, XtRGroupList, CvtStringToGroupList,
		       NULL, 0, XtCacheAll, FreeGroupList);
}

static void Initialize(Widget grequest, Widget gnew,
		       ArgList args, Cardinal *num_args)
{
    NntpObject	new = (NntpObject)gnew;
    char	buffer[1024];
    char	*home = getenv("HOME");

    if (new->nntp_object.nntp_server)
	new->nntp_object.nntp_server =
	    XtNewString(new->nntp_object.nntp_server);
    else
	XtAppError(XtWidgetToApplicationContext((Widget)new),
		   "Error: NULL nntp_server!");

    if (new->nntp_object.newsrc_file) {
	if (new->nntp_object.newsrc_file[0] == '/')
	    new->nntp_object.newsrc_file =
		XtNewString(new->nntp_object.newsrc_file);
	else if (new->nntp_object.newsrc_dir[0] == '/') {
	    sprintf(buffer, "%s/%s", new->nntp_object.newsrc_dir,
		    new->nntp_object.newsrc_file);
	    new->nntp_object.newsrc_file = XtNewString(buffer);
	} else if (home) {
	    sprintf(buffer, "%s/%s/%s", home,
		    new->nntp_object.newsrc_dir, new->nntp_object.newsrc_file);
	    new->nntp_object.newsrc_file = XtNewString(buffer);
	} else {
	    new->nntp_object.newsrc_file = NULL;
	}
    } else {
	if (new->nntp_object.newsrc_dir[0] == '/') {
	    sprintf(buffer, "%s/.newsrc-%s",
		    new->nntp_object.newsrc_dir,
		    new->nntp_object.nntp_server);
	    new->nntp_object.newsrc_file = XtNewString(buffer);
	} else if (home) {
	    sprintf(buffer, "%s/%s/.newsrc-%s", home,
		    new->nntp_object.newsrc_dir,
		    new->nntp_object.nntp_server);
	    new->nntp_object.newsrc_file = XtNewString(buffer);
	}
    }

    if (new->nntp_object.old_newsrc_file) {
	if (new->nntp_object.old_newsrc_file[0] == '/') {
	    new->nntp_object.old_newsrc_file =
		XtNewString(new->nntp_object.old_newsrc_file);
	} else if (new->nntp_object.newsrc_dir[0] == '/') {
	    sprintf(buffer, "%s/%s", new->nntp_object.newsrc_dir,
		    new->nntp_object.old_newsrc_file);
	    new->nntp_object.old_newsrc_file = XtNewString(buffer);
	} else if (home) {
	    sprintf(buffer, "%s/%s/%s", home,
		    new->nntp_object.newsrc_dir,
		    new->nntp_object.old_newsrc_file);
	    new->nntp_object.old_newsrc_file = XtNewString(buffer);
	} else {
	    new->nntp_object.old_newsrc_file = NULL;
	}
    } else {
	if (new->nntp_object.newsrc_dir[0] == '/') {
	    sprintf(buffer, "%s/.oldnewsrc-%s",
		    new->nntp_object.newsrc_dir,
		    new->nntp_object.nntp_server);
	    new->nntp_object.old_newsrc_file = XtNewString(buffer);
	} else if (home) {
	    sprintf(buffer, "%s/%s/.oldnewsrc-%s", home,
		    new->nntp_object.newsrc_dir,
		    new->nntp_object.nntp_server);
	    new->nntp_object.old_newsrc_file = XtNewString(buffer);
	}
    }

    if (new->nntp_object.kill_file) {
	if (new->nntp_object.kill_file[0] == '/') {
	    new->nntp_object.kill_file =
		XtNewString(new->nntp_object.kill_file);
	} else if (new->nntp_object.kill_dir[0] == '/') {
	    sprintf(buffer, "%s/%s", new->nntp_object.kill_dir,
		    new->nntp_object.kill_file);
	    new->nntp_object.kill_file = XtNewString(buffer);
	} else if (home) {
	    sprintf(buffer, "%s/%s/%s", home,
		    new->nntp_object.kill_dir,
		    new->nntp_object.kill_file);
	    new->nntp_object.kill_file = XtNewString(buffer);
	} else {
	    new->nntp_object.kill_file = NULL;
	}
    } else {
	if (new->nntp_object.kill_dir[0] == '/') {
	    sprintf(buffer, "%s/.kill-%s",
		    new->nntp_object.kill_dir,
		    new->nntp_object.nntp_server);
	    new->nntp_object.kill_file = XtNewString(buffer);
	} else if (home) {
	    sprintf(buffer, "%s/%s/.kill-%s", home,
		    new->nntp_object.kill_dir,
		    new->nntp_object.nntp_server);
	    new->nntp_object.kill_file = XtNewString(buffer);
	}
    }

    if (new->nntp_object.auth_info_generic) {
	if (new->nntp_object.auth_info_generic[0] != '\0') {
	    new->nntp_object.auth_info_generic =
		XtNewString(new->nntp_object.auth_info_generic);
	} else {
	    new->nntp_object.auth_info_generic = NULL;
	}
    } else {
	char	*env_var;

	env_var = getenv("NNTPAUTH");
	if (env_var && env_var[0] != '\0')
	    new->nntp_object.auth_info_generic = XtNewString(env_var);
    }
}

static void Destroy(Widget gw)
{
    NntpObject	w = (NntpObject)gw;

    XtFree(w->nntp_object.newsrc_file);
    XtFree(w->nntp_object.kill_file);
    XtFree(w->nntp_object.auth_info_generic);
}

static void Realize(Widget w, XtValueMask *mask,
		    XSetWindowAttributes *attributes)
{
    XtAppWarning(XtWidgetToApplicationContext(w),
		 "Widgets of NntpObject should never be realized.");
}

static Boolean SetValues(Widget gcurrent,
			 Widget grequest,
			 Widget gnew,
			 ArgList args,
			 Cardinal *num_args)
{
    NntpObject	new = (NntpObject)gnew;
    NntpObject	current = (NntpObject)gcurrent;

    new->nntp_object.nntp_server = current->nntp_object.nntp_server;

    return False;
}

/*************************************************************************/

char *NntpObjectKillFile(NntpObject w)
{
    if (w)
	return w->nntp_object.kill_file;
    else
	return NULL;
}

char *NntpObjectNewsrcFile(NntpObject w)
{
    if (w)
	return w->nntp_object.newsrc_file;
    else
	return NULL;
}

char *NntpObjectOldNewsrcFile(NntpObject w)
{
    if (w)
	return w->nntp_object.old_newsrc_file;
    else
	return NULL;
}

char **NntpObjectAutoSubscribe(NntpObject w)
{
    if (w)
	return w->nntp_object.auto_subscribe;
    else
	return NULL;
}

char *NntpObjectAuthInfoUser(NntpObject w)
{
    if (w)
	return w->nntp_object.auth_info_user;
    else
	return NULL;
}

char *NntpObjectAuthInfoPass(NntpObject w)
{
    if (w)
	return w->nntp_object.auth_info_pass;
    else
	return NULL;
}

char *NntpObjectAuthInfoGeneric(NntpObject w)
{
    if (w)
	return w->nntp_object.auth_info_generic;
    else
	return NULL;
}

char *NntpObjectPostingAgent(NntpObject w)
{
    if (w)
	return w->nntp_object.posting_agent;
    else
	return NULL;
}

int NntpObjectRescanTimeout(NntpObject w)
{
    if (w)
	return w->nntp_object.rescan_timeout;
    else
	return 0;
}

Boolean NntpObjectRetrieveDescriptions(NntpObject w)
{
    if (w)
	return w->nntp_object.retrieve_descriptions;
    else
	return True;
}

Boolean NntpObjectCreateNewsrc(NntpObject w)
{
    if (w)
	return w->nntp_object.create_newsrc;
    else
	return False;
}

Boolean NntpObjectCheckForNewGroups(NntpObject w)
{
    if (w)
	return w->nntp_object.check_for_new_groups;
    else
	return False;
}

Boolean NntpObjectReadActiveFile(NntpObject w)
{
    if (w)
	return w->nntp_object.read_active_file;
    else
	return True;
}

Boolean NntpObjectFillNewsrcFile(NntpObject w)
{
    if (w)
	return w->nntp_object.fill_newsrc_file;
    else
	return False;
}

Boolean NntpObjectThreadEmAll(NntpObject w)
{
    if (w)
	return w->nntp_object.thread_em_all;
    else
	return False;
}

NntpObject NntpObjectCreate(Widget parent, char *nntp_server)
{
    char	name[1024];
    Arg		arg;
    char	*src, *dest;

    sprintf(name, "%s", nntp_server);
    src = dest = name;
    for (;;) {
	if (*src == ' ' || *src == '\t') {
	    src++;
	    continue;
	}
	if (*src == '.') {
	    *dest++ = '_';
	    src++;
	    continue;
	}
	if (*src == ':') {
	    *dest++ = '_';
	    src++;
	    for (;;) {
		if (*src == '\0')
		    break;
		else if (*src >= '0' && *src <= '9')
		    *dest++ = *src++;
		else
		    src++;
	    }
	    break;
	}

	if ((*dest++ = *src++) == '\0')
	    break;
    }
    *dest = '\0';

    arg.name = XtNnntpServer;
    arg.value = (XtArgVal)nntp_server;

    /* XtNewString(name) */
    return (NntpObject)XtCreateWidget(name, nntpObjectClass,
				      parent, &arg, 1);
}
