/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#define KNEWS_VERSION		"0.9.3"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <regex.h>
#include "../Widgets/ArtTreeNode.h"
#include "NntpObject.h"
#include "GroupObjec.h"

typedef enum {
    NewsModeDisconnected,
    NewsModeConnected,
    NewsModeGroup,
    NewsModeThread,
    NewsModeAllgroups,
    NewsModeNewgroups
} NewsMode;

typedef struct subj_node {
    struct subj_node	*hash_next;
    struct subj_node	*next;
    struct subj_node	*prev;
    struct art_node	*thread;
    char		*subject;
    long		hash_len;
    long		disp;
    long		no_unread;
    short		hot;
    char		has_tagged;
} SUBJECT;

typedef struct art_node {
    ART_TREE_NODE	tree_data;
    struct art_node     *hash_next;
    struct art_node     *next;
    char                *msgid;
    long                hash_len;
    SUBJECT             *subject;
    char                *from;
    char                *xref;
    time_t		date;
    long                no;
    long		lines;
/*    long		bytes;*/
    short               hot;
    char		read;
    char		killed;
} ARTICLE;

#define PARENT(art)	((art)->tree_data.parent)
#define CHILD1(art)	((art)->tree_data.child1)
#define SIBLING(art)	((art)->tree_data.sibling)

#define A_PARENT(art)	((ARTICLE *)PARENT(art))
#define A_CHILD1(art)	((ARTICLE *)CHILD1(art))
#define A_SIBLING(art)	((ARTICLE *)SIBLING(art))

typedef enum {
    KillFieldMsgid,
    KillFieldSubject,
    KillFieldFrom,
    KillFieldXref
} KillField;

typedef enum {
    KillScopeArticle,
    KillScopeSubject,
    KillScopeThread,
    KillScopeSubthread
} KillScope;

typedef struct {
    char		*expr_str;
    regex_t		*expr_re;
    char		*group_str;
    regex_t		*group_re;
    char		*color;
    Pixel		pixel;
    short		hot;
    KillField		field;
    KillScope		scope;
    unsigned int	kill : 1,
			valid : 1,
			applicable : 1,
			unused : 1,
			alloced_pixel : 1,
			msgid_valid : 1;
} KILL_NODE;

#define REGEXP_COMPILE_FLAGS (REG_ICASE | REG_NOSUB | REG_EXTENDED)

typedef struct art_list_node {
    long                  first;
    long                  last;
    struct art_list_node  *next;
} ART_LIST_NODE;

#define HASH_SIZE  237

typedef enum {
    AheadFlagScheduled,
    AheadFlagThreading,
    AheadFlagDone
} AheadFlag;

typedef struct {
    struct group_node	*group;
    ARTICLE		*articles;
    SUBJECT		*subjects;
    ARTICLE		*art_ht[HASH_SIZE];
    SUBJECT		*subj_ht[HASH_SIZE];
    ARTICLE		*last_art;
    SUBJECT		*last_subj;
    SUBJECT		*fake_had_subject;
    char		*refs;
    long		total;
    long		n_done;
    AheadFlag		flag;
} THREAD_CONTEXT;

typedef struct group_node {
    char		*name;
    char		*description;
    THREAD_CONTEXT	*ahead;
    long		no_unread;
    long		first_art;
    long		last_art;
    ART_LIST_NODE	*read_arts;
    long		disp;
    char		subscribed;
    char		moderated;
    char		found_in_newsrc;
} GROUP_NODE;

typedef struct {
    int		fd;
    char	*buffer;
    long	curr_pos;
    long	len;
    FILE	*cookie;
} SERVER;

extern struct Global {
    String		nntp_server;
    String		edit_command;
    Cursor		cursor;
    Cursor		busy_cursor;
    String		version;
    String		mail_name;
    long		stderr_timeout;
    long		ahead_timeout;
    int			chunk_size;
    Boolean		separate_windows;
    Boolean		bell;
    Boolean		head_debug;
    /* private */
    XtAppContext	app_cont;
    SERVER		server;
    NntpObject		nntp_object;
    GroupObject		group_object;
    char		*user_id;
    char		*domain_name;
    GROUP_NODE		**groups;
    long		no_groups;
    long		max_groups;
    GROUP_NODE		**new_groups;
    long		no_new_groups;
    SUBJECT		*curr_subj;
    ARTICLE		*curr_art;
    THREAD_CONTEXT	thr;
    KILL_NODE		**kill_list;
    int			no_alloc_kill_list;
    Pixmap		*hot_pixmaps;
    int			n_pixmaps;
    time_t		last_time;
    NewsMode		mode;
    Boolean		aborted;
    Boolean		busy;
    Boolean		quit_nicely;
    Boolean		posting_allowed;
    Boolean		xover_supported;
} global;
