/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include "hash.h"

static long str_hash(unsigned char *str,	/* % doesn't work with	*/
		     long len)			/* negative numbers	*/
{
    long	hash = 0;

    while (len-- > 0)
	hash += *str++;

    return hash % HASH_SIZE;
}

void art_hash_table_clear(ARTICLE **art_ht) {
    ARTICLE	*loop, *next;
    long	i;

    for (i = 0 ; i < HASH_SIZE ; i++) {
	for (loop = art_ht[i] ; loop ; loop = next) {
	    next = loop->hash_next;
	    XtFree(loop->msgid);
	    XtFree(loop->xref);
	    XtFree(loop->from);
	    XtFree(loop->tree_data.label);
	    XtFree((char *)loop);
	}
	art_ht[i] = NULL;
    }
}

ARTICLE **art_hash_find(ARTICLE **art_ht, char *msgid, long len)
{
    ARTICLE	*loop, *prev = NULL;
    ARTICLE	**hash_ptr;

    hash_ptr = &art_ht[str_hash((unsigned char *)msgid, len)];
    for (loop = *hash_ptr ; loop ; prev = loop, loop = loop->hash_next)
	if (loop->hash_len == len &&
	    strncmp(msgid, loop->msgid, len) == 0) break;

    return prev ? &prev->hash_next : hash_ptr;
}

void subj_hash_table_clear(SUBJECT **subj_ht) {
    SUBJECT	*loop, *next;
    long	i;

    for (i = 0 ; i < HASH_SIZE ; i++) {
	for (loop = subj_ht[i] ; loop ; loop = next) {
	    next = loop->hash_next;
	    XtFree(loop->subject);
	    XtFree((char *)loop);
	}
	subj_ht[i] = NULL;
    }
}

SUBJECT **subj_hash_find(SUBJECT **subj_ht, char *subject, long len)
{
    SUBJECT	*loop, *prev = NULL;
    SUBJECT	**hash_ptr;

    hash_ptr = &subj_ht[str_hash((unsigned char *)subject, len)];
    for (loop = *hash_ptr ; loop ; prev = loop, loop = loop->hash_next)
	if (loop->hash_len == len &&
	    strncmp(subject, loop->subject, len) == 0) break;

    return prev ? &prev->hash_next : hash_ptr;
}

void clear_thread_context(THREAD_CONTEXT *context)
{
    context->articles = NULL;
    context->subjects = NULL;
    art_hash_table_clear(context->art_ht);
    subj_hash_table_clear(context->subj_ht);
    context->last_art = NULL;
    context->last_subj = NULL;
    context->fake_had_subject = NULL;
    context->refs = NULL;
    context->total = 1;
    context->n_done = 0;
    context->flag = AheadFlagScheduled;
}

void nullify_thread_context(THREAD_CONTEXT *context)
{
    int i;

    context->articles = NULL;
    context->subjects = NULL;
    context->last_art = NULL;
    context->last_subj = NULL;
    context->fake_had_subject = NULL;
    context->refs = NULL;
    context->total = 1;
    context->n_done = 0;
    context->flag = AheadFlagScheduled;

    for (i = 0 ; i < HASH_SIZE ; i++) {
	context->art_ht[i] = NULL;
	context->subj_ht[i] = NULL;
    }
}
