/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <ctype.h>
#include <X11/Shell.h>
#include "kill.h"
#include "pixmaps.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/Knapp.h"
#include "../Widgets/Layout.h"
#include "../Widgets/Menu.h"
#include "../Widgets/MenuKnapp.h"
#include "../Widgets/Message.h"
#include "../Widgets/Notice.h"
#include "../Widgets/PopdownSh.h"
#include "../Widgets/Sash.h"
#include "../Widgets/ScrBar.h"
#include "../Widgets/ScrList.h"
#include "../Widgets/SeparatorG.h"
#include "../Widgets/StringG.h"
#include "../Widgets/TextField.h"
#include "../Widgets/Util.h"

#define LINE_LEN 128

#define SAME_FIELD_TYPE(f1, f2) \
( (((f1) == KillFieldMsgid) && ((f2 == KillFieldMsgid)) ) || \
  (!((f1 == KillFieldMsgid)) && !((f2) == KillFieldMsgid)) )

static struct {
    Widget	list;
    Widget	scrbar;
    Widget	field_knapp;
    Widget	field_shell;
    Widget	field_menu;
    Widget	scope_knapp;
    Widget	scope_shell;
    Widget	scope_menu;
    Widget	kill_knapp;
    Widget	kill_shell;
    Widget	kill_menu;
    Widget	update_knapp;
    Widget	add_knapp;
    Widget	delete_knapp;
    Widget	group_field;
    Widget	expr_field;
    Widget	color_field;
    Widget	messageid_gadget;
    Widget	subject_gadget;
    Widget	from_gadget;
    Widget	xref_gadget;
} kill_widgets;

static KillField	temp_field = KillFieldSubject;
static KillScope	temp_scope = KillScopeArticle;
static Boolean		temp_kill = True;

/****************************************************************************/

static void fix_pixmap(KILL_NODE *node)
{
    Display	*disp = XtDisplay(main_widgets.kill_editor_shell);

    if (node->hot > 0) {
	Window		win = XtWindow(main_widgets.kill_editor_shell);
	GC		gc;
	XGCValues	values;

	if (node->alloced_pixel)
	    values.foreground = node->pixel;
	else
	    values.foreground = GroupObjectDefaultHotPixel(global.group_object);
	gc = XCreateGC(disp, win, GCForeground, &values);
	XFillRectangle(disp, global.hot_pixmaps[node->hot], gc,
		       0, 0, HOT_PIXMAP_SIZE, HOT_PIXMAP_SIZE);
	XFreeGC(disp, gc);
    } else {
	create_hot_pixmap(node);
    }
}

static void check_if_valid(KILL_NODE *node)
{
    if (node->expr_str && node->group_str && node->group_re &&
	((node->field == KillFieldMsgid && node->msgid_valid) ||
	 node->expr_re) &&
	(node->kill || node->color))
	node->valid = True;
    else
	node->valid = False;
}

static void print_kill_info(char *buffer, KILL_NODE *node)
{
    memset(buffer, ' ', LINE_LEN);

    buffer[0] = node->valid ? ' ' : '-';

    switch (node->field) {
    case KillFieldMsgid:
	buffer[1] = 'M';
	break;
    case KillFieldSubject:
	buffer[1] = 'S';
	break;
    case KillFieldFrom:
	buffer[1] = 'F';
	break;
    case KillFieldXref:
	buffer[1] = 'X';
	break;
    };

    switch (node->scope) {
    case KillScopeArticle:
	buffer[2] = 'A';
	break;
    case KillScopeSubject:
	buffer[2] = 'S';
	break;
    case KillScopeThread:
	buffer[2] = 'T';
	break;
    case KillScopeSubthread:
	buffer[2] = 't';
	break;
    }

    buffer[3] = node->kill ? 'K' : 'H';

    if (node->group_str) {
	int	n = strlen(node->group_str);

	if (n > 42) n = 42;
	strncpy(&buffer[5], node->group_str, n);
    }

    if (node->expr_str) {
	int	n = strlen(node->expr_str);

	if (n > LINE_LEN - 44) n = LINE_LEN - 44;
	strncpy(&buffer[44], node->expr_str, n);
    }

    buffer[LINE_LEN - 1] = '\0';
}

static void update_list_entry(long n)
{
    if (n >= 0) {
	char	buffer[LINE_LEN];
	Pixmap	pixmap;

	print_kill_info(buffer, global.kill_list[n]);
	pixmap = global.hot_pixmaps[global.kill_list[n]->hot];
	ScrListSetLine(kill_widgets.list, n, buffer, pixmap);
    }
}

static void update_field_knapp(KillField field)
{
    String	label = NULL;

    temp_field = field;
    switch (field) {
    case KillFieldMsgid:
	label = "Message-ID";
	break;
    case KillFieldSubject:
	label = "Subject";
	break;
    case KillFieldFrom:
	label = "From";
	break;
    case KillFieldXref:
	label = "Xref";
	break;
    }

    KnappSetLabel(kill_widgets.field_knapp, label, True);
}

static void update_field_menu_sensitives(KillField field, Boolean all_sensitive)
{
    XtSetSensitive(kill_widgets.messageid_gadget,
		   all_sensitive || field == KillFieldMsgid);
    XtSetSensitive(kill_widgets.subject_gadget,
		   all_sensitive || field != KillFieldMsgid);
    XtSetSensitive(kill_widgets.from_gadget,
		   all_sensitive || field != KillFieldMsgid);
    XtSetSensitive(kill_widgets.xref_gadget,
		   all_sensitive || field != KillFieldMsgid);
}

static void update_scope_knapp(KillScope scope)
{
    String	label = NULL;

    temp_scope = scope;
    switch (scope) {
    case KillScopeArticle:
	label = "Article";
	break;
    case KillScopeSubject:
	label = "Subject";
	break;
    case KillScopeThread:
	label = "Thread";
	break;
    case KillScopeSubthread:
	label = "Subthread";
	break;
    }

    KnappSetLabel(kill_widgets.scope_knapp, label, True);
}

static void update_kill_knapp(int kill)
{
    temp_kill = kill;
    KnappSetLabel(kill_widgets.kill_knapp, kill ? "Kill" : "Hot ", True);
}

static void update_kill_controls(long n)
{
    Arg		arg;

    if (n >= 0) {
	KILL_NODE	*node = global.kill_list[n];

	update_field_knapp(node->field);
	update_field_menu_sensitives(node->field, False);
	update_scope_knapp(node->scope);
	update_kill_knapp(node->kill);

	XtSetArg(arg, XtNbuffer, node->group_str);
	XtSetValues(kill_widgets.group_field, &arg, 1);
	XtSetArg(arg, XtNbuffer, node->expr_str);
	XtSetValues(kill_widgets.expr_field, &arg, 1);
	XtSetArg(arg, XtNbuffer, node->color);
	XtSetValues(kill_widgets.color_field, &arg, 1);
    } else {
	update_field_knapp(KillFieldSubject);
	update_field_menu_sensitives(KillFieldSubject, True);
	update_scope_knapp(KillScopeArticle);
	update_kill_knapp(True);

	XtSetArg(arg, XtNbuffer, NULL);
	XtSetValues(kill_widgets.group_field, &arg, 1);
	XtSetArg(arg, XtNbuffer, NULL);
	XtSetValues(kill_widgets.expr_field, &arg, 1);
	XtSetArg(arg, XtNbuffer, NULL);
	XtSetValues(kill_widgets.color_field, &arg, 1);
    }
}

static void init_kill_editor()
{
    Arg		args[8];
    long	shown, n;

    ScrListClearLines(kill_widgets.list);

    for (n = 0 ; n < global.no_alloc_kill_list ; n++) {
	KILL_NODE	*node = global.kill_list[n];
	char		buffer[LINE_LEN];
	Pixmap		pixmap;

	if (!node) break;

	print_kill_info(buffer, node);
	pixmap = node->kill ?  None : global.hot_pixmaps[node->hot];
	ScrListAddLine(kill_widgets.list, buffer, pixmap);
    }

    ScrListGetFirstShownSize(kill_widgets.list, NULL, &shown, &n);
    ScrBarSetLengthsAndPos(kill_widgets.scrbar, n, shown, 0);

    update_kill_controls(-1);
    XtSetSensitive(kill_widgets.color_field, False);
    XtSetSensitive(kill_widgets.add_knapp, True);
    XtSetSensitive(kill_widgets.delete_knapp, False);

    if (global.thr.group) {
	char	*temp;

	temp = regexp_escape_string(global.thr.group->name, True);
	XtSetArg(args[0], XtNbuffer, temp);
	XtSetValues(kill_widgets.group_field, args, 1);
	XtFree(temp);
    }

}

static void close_knapp_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    XtPopdown(main_widgets.kill_editor_shell);
}

static void clear_knapp_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    long	n = ScrListGetFirstSelected(kill_widgets.list);

    if (n >= 0) ScrListSetSelected(kill_widgets.list, n, False);
    update_kill_controls(-1);
    XtSetSensitive(kill_widgets.add_knapp, True);
    XtSetSensitive(kill_widgets.delete_knapp, False);
    XtSetKeyboardFocus(main_widgets.kill_editor_shell,
		       kill_widgets.expr_field);
    XtSetSensitive(kill_widgets.color_field, False);
}

static void delete_knapp_callback(Widget w,
				  XtPointer client_data,
				  XtPointer call_data)
{
    long	n = ScrListGetFirstSelected(kill_widgets.list);

    if (n >= 0) {
	Display		*disp = XtDisplay(main_widgets.kill_editor_shell);
	KILL_NODE	*node = global.kill_list[n];
	Arg		arg;
	long		i, shown;

	ScrListDeleteLine(kill_widgets.list, n);

	for (i = n ; i < global.no_alloc_kill_list - 1 ; i++)
	    global.kill_list[i] = global.kill_list[i + 1];
	global.kill_list[i] = NULL;

	if (node->alloced_pixel) {
	    Colormap	cmap;

	    XtSetArg(arg, XtNcolormap, &cmap);
	    XtGetValues(main_widgets.shell, &arg, 1);

	    XFreeColors(disp, cmap, &node->pixel, 1, 0);
	}

	if (node->hot > 0) {
	    Pixmap	pixmap = global.hot_pixmaps[node->hot];

	    purge_hot(node->hot);
	    global.hot_pixmaps[node->hot] = None;
	    if (pixmap != None) XFreePixmap(disp, pixmap);
	}

	if (node->expr_re) {
	    regfree(node->expr_re);
	    XtFree((char *)node->expr_re);
	    node->expr_re = NULL;
	}
	if (node->group_re) {
	    regfree(node->group_re);
	    XtFree((char *)node->group_re);
	    node->group_re = NULL;
	}
	XtFree(node->expr_str);
	XtFree(node->group_str);
	XtFree(node->color);
	XtFree((char *)node);

	ScrListGetFirstShownSize(kill_widgets.list, &i, &shown, &n);
	ScrBarSetLengthsAndPos(kill_widgets.scrbar, n, shown, i);

	XtSetSensitive(kill_widgets.add_knapp, True);
    }
}

static void add_knapp_callback(Widget w,
			       XtPointer client_data,
			       XtPointer call_data)
{
    long	n = ScrListGetFirstSelected(kill_widgets.list);

    if (n < 0) {
	KILL_NODE	*node = (KILL_NODE *)XtMalloc(sizeof(KILL_NODE));
	long		i, shown, first;
	Arg		args[3];
	char		*c;
	Boolean		valid = True;
	regex_t		re;
	char		buffer[LINE_LEN];
	Pixmap		pixmap;

	for (i = 0 ; i < global.no_alloc_kill_list ; i++)
	    if (!global.kill_list[i]) break;

	if (i + 3 > global.no_alloc_kill_list) {
	    int		j = global.no_alloc_kill_list;

	    global.no_alloc_kill_list = 2 * (global.no_alloc_kill_list + 1);
	    global.kill_list =
		(KILL_NODE **)XtRealloc((char *)global.kill_list,
					global.no_alloc_kill_list *
					sizeof(KILL_NODE *));
	    while (j < global.no_alloc_kill_list)
		global.kill_list[j++] = NULL;
	}

	global.kill_list[i] = node;

	node->applicable = False;
	node->unused = False;
	node->expr_re = NULL;
	node->group_re = NULL;
	node->field = temp_field;
	node->scope = temp_scope;

	XtSetArg(args[0], XtNbuffer, &c);
	XtGetValues(kill_widgets.group_field, args, 1);
	if (c && c[0] != '\0') {
	    node->group_str = XtNewString(c);
	    if (regcomp(&re, c, REGEXP_COMPILE_FLAGS) == 0) {
		node->group_re = (regex_t *)XtMalloc(sizeof(regex_t));
		memcpy(node->group_re, &re, sizeof(regex_t));
	    } else {
		regfree(&re);
		valid = False;
	    }
	} else {
	    node->group_str = NULL;
	    valid = False;
	}

	XtSetArg(args[0], XtNbuffer, &c);
	XtGetValues(kill_widgets.expr_field, args, 1);
	if (temp_field == KillFieldMsgid) {
	    node->msgid_valid = False;

	    if (c) {
		while (isspace((unsigned char)*c))
		    c++;
		node->expr_str = XtNewString(c);
		if (*c == '<') {
		    char	*p = c + strlen(c) - 1;

		    while (isspace((unsigned char)*p))
			p--;
		    if (*p == '>')
			node->msgid_valid = True;
		}
	    }

	    if (!node->msgid_valid) {
		node->expr_str = NULL;
		node->msgid_valid = False;
		valid = False;
	    }
	} else {
	    node->msgid_valid = False;
	    if (c && c[0] != '\0') {
		node->expr_str = XtNewString(c);
		if (regcomp(&re, c, REGEXP_COMPILE_FLAGS) == 0) {
		    node->expr_re = (regex_t *)XtMalloc(sizeof(regex_t));
		    memcpy(node->expr_re, &re, sizeof(regex_t));
		} else {
		    regfree(&re);
		    valid = False;
		}
	    } else {
		node->expr_str = NULL;
		valid = False;
	    }
	}

	node->kill = temp_kill;
	if (temp_kill) {
	    node->hot = 0;
	    node->alloced_pixel = False;
	    node->color = NULL;
	} else {
	    Display	*disp = XtDisplay(main_widgets.kill_editor_shell);
	    Colormap	cmap;
	    XColor	col;
	    int		ok;

	    XtSetArg(args[0], XtNbuffer, &c);
	    XtSetArg(args[1], XtNcolormap, &cmap);
	    XtGetValues(kill_widgets.color_field, args, 2);
	    if (!c) c = "";
	    node->color = XtNewString(c);

	    ok = XParseColor(disp, cmap, node->color, &col);
	    if (ok && XAllocColor(disp, cmap, &col)) {
		node->pixel = col.pixel;
		node->alloced_pixel = True;
	    } else {
		node->pixel =
		    GroupObjectDefaultHotPixel(global.group_object);
		node->alloced_pixel = False;
		if (ok)
		    NoticeCreateAndPopup("colornotice", w,
					 "Warning: Cannot allocate "
					 "colormap entry.",
					 "OK", NULL, NULL, 2000,
					 NULL, NULL, XtGrabExclusive);
		else {
		    if (global.bell)
			XBell(XtDisplay(w), 0);
		    NoticeCreateAndPopup("colornotice", w,
					 "Error: No such color.",
					 "OK", NULL, NULL, 2000,
					 NULL, NULL, XtGrabExclusive);
		}
	    }

	    create_hot_pixmap(node);
	}

	node->valid = valid;
	if (!valid && global.bell)
	    XBell(XtDisplay(w), 0);

	print_kill_info(buffer, global.kill_list[i]);
	if (global.kill_list[i]->kill)
	    pixmap = None;
	else
	    pixmap = global.hot_pixmaps[global.kill_list[i]->hot];
	ScrListAddLine(kill_widgets.list, buffer, pixmap);

	ScrListGetFirstShownSize(kill_widgets.list, &first, &shown, &i);
	ScrBarSetLengthsAndPos(kill_widgets.scrbar, i, shown, first);
    }
}

static void update_knapp_callback(Widget w,
				  XtPointer client_date,
				  XtPointer call_data)
{
    if (update_kill_file())
	set_message("Kill file updated.", False);
    else
	set_message("Failed to update kill file!", True);
}

static void list_sel_callback(Widget w,
			      XtPointer client_data,
			      XtPointer call_data)
{
    long	n = ScrListGetFirstSelected(kill_widgets.list);

    if (n >= 0) {
	Boolean	flag = !global.kill_list[n]->kill;

	update_kill_controls(n);
	XtSetSensitive(kill_widgets.add_knapp, False);
	XtSetSensitive(kill_widgets.delete_knapp, True);
#if (XtSpecificationRelease >= 6)
	if (!flag &&
	    kill_widgets.color_field ==
	    XtGetKeyboardFocusWidget(main_widgets.kill_editor_shell)) {
	    XtSetKeyboardFocus(main_widgets.kill_editor_shell,
			       kill_widgets.expr_field);
	}
#endif
	XtSetSensitive(kill_widgets.color_field, flag);
    } else {
	XtSetSensitive(kill_widgets.add_knapp, True);
	XtSetSensitive(kill_widgets.delete_knapp, False);
    }
}

static void list_dnd_callback(Widget w,
			      XtPointer client_data,
			      XtPointer call_data)
{
    long	*index = (long *)call_data;
    long	n;
    KILL_NODE	*temp;

    index[2] = False;

    if (global.busy || !global.kill_list) return;

    for (n = 0 ; global.kill_list[n] ; n++);
    if (index[0] >= n || index[1] >= n) return;

    index[2] = True;
    temp = global.kill_list[index[0]];
    if (index[0] < index[1])
	for (n = index[0] ; n < index[1] ; n++)
	    global.kill_list[n] = global.kill_list[n + 1];
    else
	for (n = index[0] ; n > index[1] ; n--)
	    global.kill_list[n] = global.kill_list[n - 1];
    global.kill_list[index[1]] = temp;
}

static void list_resize_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    ScrListReport	*report = (ScrListReport *)call_data;

    ScrBarSetLengthsAndPos(kill_widgets.scrbar, report->canvas_length,
			   report->slider_length, report->slider_position);
}

static void scrbar_callback(Widget w,
			    XtPointer client_data,
			    XtPointer call_data)
{
    ScrBarReport	*report = (ScrBarReport *)call_data;
    long		shown, size;

    ScrListSetFirst(kill_widgets.list, report->slider_position);
    ScrListGetFirstShownSize(kill_widgets.list, NULL, &shown, &size);
    if ((report->slider_length != shown || report->canvas_length != size) &&
	report->slider_position + shown < size)
	ScrBarSetLengthsAndPos(kill_widgets.scrbar,
			       size, shown, report->slider_position);
}

static void field_menu_callback(Widget w,
				XtPointer client_data,
				XtPointer call_data)
{
    KillField	field = (KillField)client_data;
    long	n = ScrListGetFirstSelected(kill_widgets.list);

    if (n >= 0) {
	KILL_NODE	*node = global.kill_list[n];

	if (SAME_FIELD_TYPE(node->field, field)) {
	    global.kill_list[n]->field = field;
	    update_list_entry(n);
	    update_field_knapp(field);
	} else {
	    if (global.bell)
		XBell(XtDisplay(w), 0);
	}
    } else {
	char	*buffer;
	Arg	arg;

	XtSetArg(arg, XtNbuffer, &buffer);
	XtGetValues(kill_widgets.expr_field, &arg, 1);

	if (buffer && buffer[0] != '\0') {
	    if (SAME_FIELD_TYPE(temp_field, field))
		update_field_knapp(field);
	    else if (global.bell)
		XBell(XtDisplay(w), 0);
	    update_field_menu_sensitives(temp_field, False);
	} else {
	    buffer = NULL;

	    switch (field) {
	    case KillFieldMsgid:
		if (global.curr_art) {
		    buffer = XtMalloc(global.curr_art->hash_len + 3);
		    sprintf(buffer, "<%s>", global.curr_art->msgid);
		}
		break;
	    case KillFieldSubject:
		if (global.curr_art)
		    buffer =
			regexp_escape_string(global.curr_art->subject->subject,
					     False);
		else if (global.curr_subj)
		    buffer =
			regexp_escape_string(global.curr_subj->subject,
					     False);
		break;
	    case KillFieldFrom:
		if (global.curr_art && global.curr_art->from)
		    buffer =
			regexp_escape_string(global.curr_art->from,
					     False);
		break;
	    case KillFieldXref:
		break;
	    }

	    XtSetArg(arg, XtNbuffer, buffer);
	    XtSetValues(kill_widgets.expr_field, &arg, 1);
	    XtFree(buffer);

	    update_field_knapp(field);
	    if (buffer)
		update_field_menu_sensitives(field, False);
	}
    }
}

static void scope_menu_callback(Widget w,
				XtPointer client_data,
				XtPointer call_data)
{
    KillScope	scope = (KillScope)client_data;
    long	n = ScrListGetFirstSelected(kill_widgets.list);

    if (n >= 0) {
	global.kill_list[n]->scope = scope;
	update_list_entry(n);
    }

    update_scope_knapp(scope);
}

static void kill_menu_callback(Widget w,
			       XtPointer client_data,
			       XtPointer call_data)
{
    long	n = ScrListGetFirstSelected(kill_widgets.list);
    Arg		arg;

    if (client_data) {
	if (n >= 0) {
	    KILL_NODE	*node = global.kill_list[n];

	    if (node->alloced_pixel) {
		Display		*disp = XtDisplay(main_widgets.
						  kill_editor_shell);
		Colormap	cmap;
		Arg		arg;

		XtSetArg(arg, XtNcolormap, &cmap);
		XtGetValues(main_widgets.kill_editor_shell, &arg, 1);

		XFreeColors(disp, cmap, &node->pixel, 1, 0);
		node->alloced_pixel = False;
	    }

	    if (node->hot > 0) {
		purge_hot(node->hot);
		node->hot = 0;
	    }

	    node->kill = True;
	    update_list_entry(n);
	}

	XtSetArg(arg, XtNbuffer, NULL);
	XtSetValues(kill_widgets.color_field, &arg, 1);
#if (XtSpecificationRelease >= 6)
	if (kill_widgets.color_field ==
	    XtGetKeyboardFocusWidget(main_widgets.kill_editor_shell)) {
	    XtSetKeyboardFocus(main_widgets.kill_editor_shell,
			       kill_widgets.expr_field);
	}
#endif
	XtSetSensitive(kill_widgets.color_field, False);
	
	update_kill_knapp(True);
    } else {
	if (n >= 0) {
	    KILL_NODE	*node = global.kill_list[n];

	    node->kill = False;
	    update_list_entry(n);
	}

	XtSetSensitive(kill_widgets.color_field, True);

	update_kill_knapp(False);
    }
}

static void group_field_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    char	*buffer = (char *)call_data;
    long	n = ScrListGetFirstSelected(kill_widgets.list);
    Boolean	valid;
    regex_t	re;

    if (!buffer) buffer = "";

    if (regcomp(&re, buffer, REGEXP_COMPILE_FLAGS) == 0)
	valid = True;
    else {
	valid = False;
	regfree(&re);
    }

    if (n >= 0) {
	KILL_NODE	*node = global.kill_list[n];

	XtFree(node->group_str);
	node->group_str = XtNewString(buffer);

	if (valid) {
	    if (node->group_re)
		regfree(node->group_re);
	    else
		node->group_re = (regex_t *)XtMalloc(sizeof(regex_t));
	    memcpy(node->group_re, &re, sizeof(regex_t));
	} else {
	    if (node->group_re) {
		regfree(node->group_re);
		XtFree((char *)node->group_re);
		node->group_re = NULL;
	    }
	}

	check_if_valid(node);
	update_list_entry(n);
    } else if (valid) {
	regfree(&re);
    }

    if (!valid) {
	if (global.bell)
	    XBell(XtDisplay(w), 0);
	(void)NoticeCreateAndPopup("groupfieldnotice", w,
				   "Parse error in regexp.",
				   "OK", NULL, NULL, 2000,
				   NULL, NULL, XtGrabExclusive);
    } else {
	XtSetKeyboardFocus(main_widgets.kill_editor_shell,
			   kill_widgets.expr_field);
    }
}

static void expr_field_callback(Widget w,
				XtPointer client_data,
				XtPointer call_data)
{
    char	*buffer = (char *)call_data;
    long	n = ScrListGetFirstSelected(kill_widgets.list);
    Arg		arg;
    Boolean	valid = False;

    /* assert SAME_FIELD_TYPE(node->field, temp_field) */
    if (!buffer) buffer = "";

    if (temp_field == KillFieldMsgid) {
	char	*c = buffer;

	while (isspace((unsigned char)*c))
	    c++;
	c = XtNewString(c);
	if (c[0] == '<') {
	    char	*p;

	    p = c + strlen(c) - 1;
	    while (isspace((unsigned char)*p))
		p--;
	    if (p[0] == '>') {
		p[1] = '\0';
		p = strchr(c, '@');
		if (p) {
		    do {
			p++;
			if ('A' <= p[0] && p[0] <= 'Z') p[0] += 'a' - 'A';
		    } while (*p != '>');
		    valid = True;
		}
	    }
	}
			
	if (n >= 0) {
	    KILL_NODE	*node = global.kill_list[n];

	    XtFree(node->expr_str);
	    node->expr_str = c;
	    if (node->expr_re)
		regfree(node->expr_re);
	    node->expr_re = NULL;
	    node->msgid_valid = valid;
	    check_if_valid(node);
	    update_list_entry(n);
	}

	if (!valid) {
	    if (global.bell)
		XBell(XtDisplay(w), 0);
	    (void)NoticeCreateAndPopup("exprfieldnotice", w,
				       "Invalid Message-ID.",
				       "OK", NULL, NULL, 2000,
				       NULL, NULL, XtGrabExclusive);
	}

	XtSetArg(arg, XtNbuffer, c);
	XtSetValues(kill_widgets.expr_field, &arg, 1);
	
	if (n < 0) XtFree(c);
    } else {
	regex_t	re;

	if (regcomp(&re, buffer, REGEXP_COMPILE_FLAGS) == 0)
	    valid = True;
	else {
	    regfree(&re);
	    valid = False;
	}

	if (n >= 0) {
	    KILL_NODE	*node = global.kill_list[n];

	    XtFree(node->expr_str);
	    node->expr_str = XtNewString(buffer);

	    if (valid) {
		if (node->expr_re) regfree(node->expr_re);
		else node->expr_re = (regex_t *)XtMalloc(sizeof(regex_t));
		memcpy(node->expr_re, &re, sizeof(regex_t));
	    } else {
		if (node->expr_re) {
		    regfree(node->expr_re);
		    XtFree((char *)node->expr_re);
		    node->expr_re = NULL;
		}
	    }

	    check_if_valid(node);
	    update_list_entry(n);
	}

	if (!valid) {
	    if (global.bell)
		XBell(XtDisplay(w), 0);
	    (void)NoticeCreateAndPopup("exprfieldnotice", w,
				       "Parse error in regexp.",
				       "OK", NULL, NULL, 2000,
				       NULL, NULL, XtGrabExclusive);
	}
    }

    if (valid) {
	if (temp_kill)
	    XtSetKeyboardFocus(main_widgets.kill_editor_shell,
			       kill_widgets.group_field);
	else
	    XtSetKeyboardFocus(main_widgets.kill_editor_shell,
			       kill_widgets.color_field);
    }

    if (buffer[0] == '\0')
	update_field_menu_sensitives(KillFieldMsgid, True);
    else
	update_field_menu_sensitives(temp_field, False);
}

static void color_field_callback(Widget w,
				 XtPointer client_data,
				 XtPointer call_data)
{
    char	*buffer = (char *)call_data;
    long	n = ScrListGetFirstSelected(kill_widgets.list);
    Display	*disp = XtDisplay(w);
    Colormap	cmap = 0;
    XColor	col;
    Arg		arg;
    int		ok;

    /* assert temp_kill is False */

    if (!buffer) buffer = "";

    XtSetArg(arg, XtNcolormap, &cmap);
    XtGetValues(w, &arg, 1);

    if (n >= 0) {
	KILL_NODE	*node = global.kill_list[n];

	if (node->alloced_pixel)
	    XFreeColors(disp, cmap, &node->pixel, 1, 0);
	node->alloced_pixel = False;
	XtFree(node->color);
	node->color = XtNewString(buffer);

	ok = XParseColor(disp, cmap, buffer, &col);
	if (ok && XAllocColor(disp, cmap, &col)) {
	    node->alloced_pixel = True;
	    node->pixel = col.pixel;
	} else {
	    if (ok)
		NoticeCreateAndPopup("colornotice", w,
				     "Warning: Cannot allocate "
				     "colormap entry.",
				     "OK", NULL, NULL, 2000,
				     NULL, NULL, XtGrabExclusive);
	    else {
		if (global.bell)
		    XBell(XtDisplay(w), 0);
		NoticeCreateAndPopup("colornotice", w,
				     "Error: No such color.",
				     "OK", NULL, NULL, 2000,
				     NULL, NULL, XtGrabExclusive);
	    }
	}

	fix_pixmap(node);
	update_list_entry(n);
    } else {
	ok = XParseColor(disp, cmap, buffer, &col);
	if (!ok) {
	    if (global.bell)
		XBell(XtDisplay(w), 0);
	    NoticeCreateAndPopup("colornotice", w,
				 "Error: No such color.",
				 "OK", NULL, NULL, 2000,
				 NULL, NULL, XtGrabExclusive);
	}
    }

    if (ok)
	XtSetKeyboardFocus(main_widgets.kill_editor_shell,
			   kill_widgets.group_field);
}

/*************************************************************************/

static void create_kill_widgets()
{
    Arg		args[8];
    Widget	layout, temp;

    XtSetArg(args[0], XtNallowShellResize, True);
    XtSetArg(args[1], XtNinput, True);
    XtSetArg(args[2], XtNtitle, "knews: kill file editor");
    main_widgets.kill_editor_shell =
	XtCreatePopupShell("killeditor", topLevelShellWidgetClass,
			   main_widgets.shell, args, 3);

    layout =
	XtCreateManagedWidget("killayout", layoutWidgetClass,
			      main_widgets.kill_editor_shell, NULL, 0);

    XtSetArg(args[0], XtNatMostOne, True);
    XtSetArg(args[1], XtNatLeastOne, False);
    XtSetArg(args[2], XtNpixmapWidth, HOT_PIXMAP_SIZE);
    XtSetArg(args[3], XtNpixmapHeight, HOT_PIXMAP_SIZE);
    XtSetArg(args[4], XtNdepthOne, False);
    XtSetArg(args[5], XtNallowDnd, True);
    XtSetArg(args[6], XtNusePixmaps, True);
    kill_widgets.list =
	XtCreateManagedWidget("killist", scrListWidgetClass,
			      layout, args, 7);
    XtAddCallback(kill_widgets.list, XtNselectCallback,
		  list_sel_callback, NULL);
    XtAddCallback(kill_widgets.list, XtNresizeCallback,
		  list_resize_callback, NULL);
    XtAddCallback(kill_widgets.list, XtNdndCallback,
		  list_dnd_callback, NULL);

    XtSetArg(args[0], XtNallowOff, True);
    kill_widgets.scrbar =
	XtCreateManagedWidget("scrbar", scrBarWidgetClass,
			      layout, args, 1);
    XtAddCallback(kill_widgets.scrbar, XtNscrollCallback,
		  scrbar_callback, NULL);

    XtSetArg(args[0], XtNcursor, None);
    (void)XtCreateManagedWidget("sash", sashWidgetClass, layout, args, 1);

    XtSetArg(args[0], XtNlabel, "Close");
    temp = XtCreateManagedWidget("close", knappWidgetClass, layout, args, 1);
    XtAddCallback(temp, XtNcallback, close_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Delete");
    kill_widgets.delete_knapp =
	XtCreateManagedWidget("delete", knappWidgetClass, layout, args, 1);
    XtAddCallback(kill_widgets.delete_knapp, XtNcallback,
		  delete_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Clear");
    temp = XtCreateManagedWidget("clear", knappWidgetClass, layout, args, 1);
    XtAddCallback(temp, XtNcallback, clear_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Add new");
    kill_widgets.add_knapp =
	XtCreateManagedWidget("add", knappWidgetClass, layout, args, 1);
    XtAddCallback(kill_widgets.add_knapp, XtNcallback,
		  add_knapp_callback, NULL);

    XtSetArg(args[0], XtNlabel, "Update");
    kill_widgets.update_knapp =
	XtCreateManagedWidget("update", knappWidgetClass,
			      layout, args, 1);
    XtAddCallback(kill_widgets.update_knapp, XtNcallback,
		  update_knapp_callback, NULL);

    XtSetArg(args[0], XtNmenuName, "fieldshell");
    XtSetArg(args[1], XtNlabel, "          ");
    XtSetArg(args[2], XtNresizable, False);
    kill_widgets.field_knapp =
	XtCreateManagedWidget("fieldknapp", menuKnappWidgetClass,
			      layout, args, 3);

    kill_widgets.field_shell =
	XtCreatePopupShell("fieldshell", popdownShellWidgetClass,
			   main_widgets.kill_editor_shell, NULL, 0);

    kill_widgets.field_menu =
	XtCreateManagedWidget("fieldmenu", menuWidgetClass,
			      kill_widgets.field_shell, NULL, 0);

    XtSetArg(args[0], XtNbuffer, "Header:");
    (void)XtCreateManagedWidget("fieldmessage", messageWidgetClass,
				layout, args, 1);
    
    XtSetArg(args[0], XtNlabel, "Message-ID");
    kill_widgets.messageid_gadget =
	MenuCreateGadget("messageid", stringGadgetClass,
			 kill_widgets.field_menu, args, 1);
    XtAddCallback(kill_widgets.messageid_gadget, XtNcallback,
		  field_menu_callback, (XtPointer)KillFieldMsgid);
    (void)MenuCreateGadget("separator", separatorGadgetClass,
			   kill_widgets.field_menu, NULL, 0);
    XtSetArg(args[0], XtNlabel, "Subject");
    kill_widgets.subject_gadget =
	MenuCreateGadget("subject", stringGadgetClass,
			 kill_widgets.field_menu, args, 1);
    XtAddCallback(kill_widgets.subject_gadget, XtNcallback,
		  field_menu_callback, (XtPointer)KillFieldSubject);
    XtSetArg(args[0], XtNlabel, "From");
    kill_widgets.from_gadget =
	MenuCreateGadget("from", stringGadgetClass,
			 kill_widgets.field_menu, args, 1);
    XtAddCallback(kill_widgets.from_gadget, XtNcallback,
		  field_menu_callback, (XtPointer)KillFieldFrom);
    XtSetArg(args[0], XtNlabel, "Xref");
    kill_widgets.xref_gadget =
	MenuCreateGadget("xref", stringGadgetClass,
			 kill_widgets.field_menu, args, 1);
    XtAddCallback(kill_widgets.xref_gadget, XtNcallback,
		  field_menu_callback, (XtPointer)KillFieldXref);

    XtSetArg(args[0], XtNmenuName, "scopeshell");
    XtSetArg(args[1], XtNlabel, "         ");
    XtSetArg(args[2], XtNresizable, False);
    kill_widgets.scope_knapp =
	XtCreateManagedWidget("scopeknapp", menuKnappWidgetClass,
			      layout, args, 3);

    kill_widgets.scope_shell =
	XtCreatePopupShell("scopeshell", popdownShellWidgetClass,
			   main_widgets.kill_editor_shell, NULL, 0);

    kill_widgets.scope_menu =
	XtCreateManagedWidget("scopemenu", menuWidgetClass,
			      kill_widgets.scope_shell, NULL, 0);

    XtSetArg(args[0], XtNbuffer, "Scope:");
    (void)XtCreateManagedWidget("scopemessage", messageWidgetClass,
				layout, args, 1);

    XtSetArg(args[0], XtNlabel, "Article");
    temp = MenuCreateGadget("article", stringGadgetClass,
			    kill_widgets.scope_menu, args, 1);
    XtAddCallback(temp, XtNcallback,
		  scope_menu_callback, (XtPointer)KillScopeArticle);
    XtSetArg(args[0], XtNlabel, "Subject");
    temp = MenuCreateGadget("subject", stringGadgetClass,
			    kill_widgets.scope_menu, args, 1);
    XtAddCallback(temp, XtNcallback,
		  scope_menu_callback, (XtPointer)KillScopeSubject);
    XtSetArg(args[0], XtNlabel, "Thread");
    temp = MenuCreateGadget("thread", stringGadgetClass,
			    kill_widgets.scope_menu, args, 1);
    XtAddCallback(temp, XtNcallback,
		  scope_menu_callback, (XtPointer)KillScopeThread);
    XtSetArg(args[0], XtNlabel, "Subthread");
    temp = MenuCreateGadget("subthread", stringGadgetClass,
			    kill_widgets.scope_menu, args, 1);
    XtAddCallback(temp, XtNcallback,
		  scope_menu_callback, (XtPointer)KillScopeSubthread);

    XtSetArg(args[0], XtNlabel, "     ");
    XtSetArg(args[1], XtNresizable, False);
    XtSetArg(args[2], XtNmenuName, "killshell");
    kill_widgets.kill_knapp =
	XtCreateManagedWidget("killknapp", menuKnappWidgetClass,
			      layout, args, 3);

    kill_widgets.kill_shell =
	XtCreatePopupShell("killshell", popdownShellWidgetClass,
			   main_widgets.kill_editor_shell, NULL, 0);

    kill_widgets.kill_menu =
	XtCreateManagedWidget("killmenu", menuWidgetClass,
			      kill_widgets.kill_shell, NULL, 0);

    XtSetArg(args[0], XtNlabel, "Kill");
    temp = MenuCreateGadget("kill", stringGadgetClass,
			    kill_widgets.kill_menu, args, 1);
    XtAddCallback(temp, XtNcallback, kill_menu_callback, (XtPointer)True);
    XtSetArg(args[0], XtNlabel, "Hot");
    temp = MenuCreateGadget("hot", stringGadgetClass,
			    kill_widgets.kill_menu, args, 1);
    XtAddCallback(temp, XtNcallback, kill_menu_callback, (XtPointer)False);

    XtSetArg(args[0], XtNfocusRoot, main_widgets.kill_editor_shell);
    kill_widgets.group_field =
	XtCreateManagedWidget("groupfield", textFieldWidgetClass,
			      layout, args, 1);
    XtAddCallback(kill_widgets.group_field, XtNcallback,
		  group_field_callback, NULL);
    kill_widgets.expr_field =
	XtCreateManagedWidget("exprfield", textFieldWidgetClass,
			      layout, args, 1);
    XtAddCallback(kill_widgets.expr_field, XtNcallback,
		  expr_field_callback, NULL);
    XtSetArg(args[1], XtNpreferredChars, 16);
    kill_widgets.color_field =
	XtCreateManagedWidget("colorfield", textFieldWidgetClass,
			      layout, args, 2);
    XtAddCallback(kill_widgets.color_field, XtNcallback,
		  color_field_callback, NULL);

    XtSetArg(args[0], XtNbuffer, "Group regexp:");
    (void)XtCreateManagedWidget("groupmessage", messageWidgetClass,
				layout, args, 1);
    XtSetArg(args[0], XtNbuffer, "Regexp/Message-ID:");
    (void)XtCreateManagedWidget("exprmessage", messageWidgetClass,
				layout, args, 1);
    XtSetArg(args[0], XtNbuffer, "Color:");
    (void)XtCreateManagedWidget("colormessage", messageWidgetClass,
				layout, args, 1);

    XtSetKeyboardFocus(main_widgets.kill_editor_shell,
		       kill_widgets.expr_field);
    XtRealizeWidget(main_widgets.kill_editor_shell);
    XtInstallAllAccelerators(main_widgets.kill_editor_shell,
			     main_widgets.kill_editor_shell);

    add_WM_DELETE_WINDOW_callback(main_widgets.kill_editor_shell,
				  close_knapp_callback, NULL);
}

static Boolean	kill_editor_created = False;

void popup_kill_editor()
{
    if (!kill_editor_created) {
	create_kill_widgets();
	kill_editor_created = True;
    }
    init_kill_editor();
    XtPopup(main_widgets.kill_editor_shell, XtGrabNone);
}

void popdown_kill_editor()
{
    if (kill_editor_created)
	XtPopdown(main_widgets.kill_editor_shell);
}
