/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include <ctype.h>
#include "hash.h"
#include "kill.h"
#include "pixmaps.h"
#include "util.h"

#define KILL_SIZE	16
#define IS_SPACE(c)	((c) == ' ' || (c) == '\t')
#define SKIP_SPACE(c)	while (IS_SPACE(*c)) c++

void parse_kill_file()
{
    FILE	*ifp_in = NULL;
    char	buffer[1024];
    char	*file_name = NntpObjectKillFile(global.nntp_object);
    KILL_NODE	**list = NULL;
    int		n, no_alloc = 0;
    regex_t	re;

    global.kill_list = NULL;

    if (file_name)
	ifp_in = fopen(file_name, "r");

    if (ifp_in) {
	for (n = 0 ; fgets(buffer, sizeof(buffer), ifp_in) ; n++) {
	    char	*c, *p;
	    
	    c = strchr(buffer, '\n');
	    if (c) *c = '\0';
	    
	    for (c = buffer ; IS_SPACE(*c) ; c++);
	    if (*c == '\0') {
		n--;
		continue;
	    }

	    if (n + 2 > no_alloc) {
		no_alloc += KILL_SIZE;
		list =
		    (KILL_NODE **)XtRealloc((char *)list,
					    no_alloc * sizeof(KILL_NODE*));
	    }

	    list[n] = (KILL_NODE *)XtMalloc(sizeof(KILL_NODE));
	    list[n]->expr_str = NULL;
	    list[n]->expr_re = NULL;
	    list[n]->group_str = NULL;
	    list[n]->group_re = NULL;
	    list[n]->color = NULL;
	    list[n]->hot = 0;
	    list[n]->field = KillFieldSubject;
	    list[n]->scope = KillScopeArticle;
	    list[n]->valid = True;
	    list[n]->alloced_pixel = False;
	    list[n]->msgid_valid = False;

	    SKIP_SPACE(c);
	    switch (*c++) {
	    case 'M':
	    case 'm':
		list[n]->field = KillFieldMsgid;
		break;
	    case 'S':
	    case 's':
		list[n]->field = KillFieldSubject;
		break;
	    case 'F':
	    case 'f':
		list[n]->field = KillFieldFrom;
		break;
	    case 'X':
	    case 'x':
		list[n]->field = KillFieldXref;
		break;
	    default:
		fprintf(stderr,
			"knews: Unknown kill field "
			"'%c' in kill file.\n",
			*c);
		list[n]->valid = False;
		break;
	    }

	    SKIP_SPACE(c);
	    switch (*c++) {
	    case 'A':
	    case 'a':
		list[n]->scope = KillScopeArticle;
		break;
	    case 'S':
	    case 's':
		list[n]->scope = KillScopeSubject;
		break;
	    case 'T':
		list[n]->scope = KillScopeThread;
		break;
	    case 't':
		list[n]->scope = KillScopeSubthread;
		break;
	    case '\0':
		fprintf(stderr,
			"knews: Premature end of line in kill file.\n");
		list[n]->valid = False;
		continue;
	    default:
		fprintf(stderr,
			"knews: Unknown kill scope "
			"'%c' in kill file.\n",
			*c);
		list[n]->valid = False;
		break;
	    }

	    SKIP_SPACE(c);
	    switch (*c++) {
	    case 'K':
	    case 'k':
		list[n]->kill = True;
		SKIP_SPACE(c);
		if (c[0] == '\0')
		    break;
		else if (c[0] == '|' && c[1] == '|')
		    c += 2;
		else {
		    fprintf(stderr,
			    "knews: Color specified for kill entry"
			    "in kill file, ignored.\n");
		    c = strstr(c, "||");
		    if (c) c += 2;
		}
		break;
	    case 'H':
	    case 'h':
		list[n]->kill = False;
		SKIP_SPACE(c);
		if (c[0] == '\0')
		    break;
		else if (c[0] == '|' && c[1] == '|') {
		    fprintf(stderr,
			    "knews: No color specified for hot entry"
			    "in kill file.\n");
		    c += 2;
		    list[n]->valid = False;
		} else {
		    p = strstr(c, "||");
		    if (p) {
			*p = '\0';
			list[n]->color = XtNewString(c);
			c = p + 2;
		    } else
			c = NULL;
		}
		break;
	    case '\0':
		break;
	    default:
		list[n]->valid = False;
		fprintf(stderr,
			"knews: Error in kill file expected 'H' or 'K'.\n");
		c = strstr(c, "||");
		break;
	    }

	    if (!c || *c == '\0') {
		fprintf(stderr,
			"knews: Premature end of line in kill file.\n");
		list[n]->valid = False;
		continue;
	    }

	    p = strstr(c, "||");
	    if (p) *p = '\0';
	    list[n]->group_str = XtNewString(c);
	    if (!p) {
		list[n]->valid = False;
		fprintf(stderr,
			"knews: Premature end of line in kill file.\n");
		continue;
	    }
	    if (regcomp(&re, c, REGEXP_COMPILE_FLAGS) == 0) {
		list[n]->group_re = (regex_t *)XtMalloc(sizeof(regex_t));
		memcpy(list[n]->group_re, &re, sizeof(regex_t));
	    } else {
		regfree(&re);
		list[n]->valid = False;
		fprintf(stderr,
			"knews: Parse error in group regexp: %s\n", c);
	    }

	    c = p + 2;
	    SKIP_SPACE(c);
	    if (*c == '\0') {
		list[n]->valid = False;
		fprintf(stderr,
			"knews: Premature end of line in kill file.\n");
		continue;
	    }

	    if (list[n]->field == KillFieldMsgid) {
		Boolean	valid = True;
		
		if (*c != '<') valid = False;
		p = c + strlen(c) - 1;
		while (IS_SPACE(*p))
		    p--;
		if (*p != '>')
		    valid = False;
		*(p+1) = '\0';

		p = strchr(c, '@');
		for (p++ ; *p != '\0' && *p != '>' ; p++)
		    if ('A' <= *p && *p <= 'Z') *c += 'a' - 'A';
		if (*p != '>')
		    valid = False;

		list[n]->expr_str = XtNewString(c);
		list[n]->msgid_valid = valid;
		if (!valid) {
		    list[n]->valid = False;
		    fprintf(stderr,
			    "knews: Invalid msgid in kill file: %s\n",
			    c);
		}
	    } else {
		list[n]->expr_str = XtNewString(c);
		if (regcomp(&re, c, REGEXP_COMPILE_FLAGS) == 0) {
		    list[n]->expr_re = (regex_t *)XtMalloc(sizeof(regex_t));
		    memcpy(list[n]->expr_re, &re, sizeof(regex_t));
		} else {
		    regfree(&re);
		    list[n]->valid = False;
		    fprintf(stderr,
			    "knews: Parse error in expression regexp: %s\n",
			    c);
		}
	    }
	}

	fclose(ifp_in);

	while (n < no_alloc)
	    list[n++] = NULL;
	global.kill_list = list;
	global.no_alloc_kill_list = no_alloc;
    }
    
    alloc_hot_pixmaps();
}

int update_kill_file()
{
    char	*file_name = NntpObjectKillFile(global.nntp_object);
    FILE	*ifp_out;
    KILL_NODE	**loop = global.kill_list;

    if (!file_name)
	return False;
    if (!loop)
	return True;
    ifp_out = fopen(file_name, "w");
    if (!ifp_out)
	return False;

    while (*loop) {
	KILL_NODE	*node = *loop;

	switch (node->field) {
	case KillFieldMsgid:
	    fprintf(ifp_out, "M");
	    break;
	case KillFieldSubject:
	    fprintf(ifp_out, "S");
	    break;
	case KillFieldFrom:
	    fprintf(ifp_out, "F");
	    break;
	case KillFieldXref:
	    fprintf(ifp_out, "X");
	    break;
	}

	switch (node->scope) {
	case KillScopeArticle:
	    fprintf(ifp_out, "A");
	    break;
	case KillScopeSubject:
	    fprintf(ifp_out, "S");
	    break;
	case KillScopeThread:
	    fprintf(ifp_out, "T");
	    break;
	case KillScopeSubthread:
	    fprintf(ifp_out, "t");
	    break;
	}

	fprintf(ifp_out, "%c", node->kill ? 'K' : 'H');
	if (node->color) fprintf(ifp_out, "%s", node->color);
	fprintf(ifp_out, "||");
	if (node->group_str) fprintf(ifp_out, "%s", node->group_str);
	fprintf(ifp_out, "||");
	if (node->expr_str) fprintf(ifp_out, "%s", node->expr_str);
	fprintf(ifp_out, "\n");

	loop++;
    }

    fclose(ifp_out);
    
    return True;
}

void free_kill_file()
{
    int	i;

    free_hot_pixmaps();

    for (i = 0 ; i < global.no_alloc_kill_list ; i++) {
	KILL_NODE	*node = global.kill_list[i];

	if (!node) break;

	XtFree(node->expr_str);
	XtFree(node->group_str);
	XtFree(node->color);
	if (node->expr_re) {
	    regfree(node->expr_re);
	    XtFree((char *)node->expr_re);
	}
	if (node->group_re) {
	    regfree(node->group_re);
	    XtFree((char *)node->group_re);
	}
	XtFree((char *)node);
    }

    XtFree((char *)global.kill_list);
    global.kill_list = NULL;
    global.no_alloc_kill_list = 0;
}

void kill_articles(long no[2])
{
    KILL_NODE	**loop = global.kill_list;
    SUBJECT	*subj;
    ARTICLE	*art;

    no[0] = 0;
    no[1] = 0;

    if (!loop) return;

    for ( ; *loop ; loop++) {
	KILL_NODE	*node = *loop;

	if (!node->valid || !node->group_re
	    || regexec(node->group_re, global.thr.group->name,
		       0, NULL, 0) != 0)
	    continue;

	switch (node->field) {
	case KillFieldMsgid:
	    if (!node->applicable)
		node->unused = True;
	    node->applicable = True;
	    art = *art_hash_find(global.thr.art_ht,
				 &node->expr_str[1],
				 strlen(node->expr_str) - 2);
	    if (art) {
		node->unused = False;
		if (node->kill) {
		    switch (node->scope) {
		    case KillScopeArticle:
			if (art->from && !art->read) {
			    art->read = True;
			    art->killed = True;
			    no[0]++;
			    global.thr.group->no_unread--;
			    art->subject->no_unread--;
			}
			break;
		    case KillScopeSubject:
			no[0] += mark_subject_read(art->subject,
						   False, True);
			break;
		    case KillScopeThread:
			no[0] += mark_thread_read(art->subject->thread,
						  False, True);
			break;
		    case KillScopeSubthread:
			no[0] += mark_subthread_read(art, False, True);
			break;
		    }
		} else { /* hot */
		    switch (node->scope) {
		    case KillScopeArticle:
			if (art->from && art->hot == 0) {
			    art->hot = node->hot;
			    if (!art->read) no[1]++;
			}
			break;
		    case KillScopeSubject:
			no[1] += mark_subject_hot(art->subject, node->hot);
			break;
		    case KillScopeThread:
			no[1] += mark_thread_hot(art->subject->thread,
						 node->hot);
			break;
		    case KillScopeSubthread:
			no[1] += mark_subthread_hot(art, node->hot);
			break;
		    }
		}
	    } else node->unused = True;
	    break;
	case KillFieldSubject:
	    switch (node->scope) {
	    case KillScopeSubject:
	    case KillScopeArticle:
		for (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    if (regexec(node->expr_re, subj->subject,
				0, NULL, 0) == 0) {
			if (node->kill)
			    no[0] += mark_subject_read(subj, False, True);
			else
			    no[1] += mark_subject_hot(subj, node->hot);
		    }
		}
		break;
	    case KillScopeThread:
		for (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    if (regexec(node->expr_re, subj->subject,
				0, NULL, 0) == 0) {
			if (node->kill)
			    no[0] += mark_thread_read(subj->thread,
						      False, True);
			else
			    no[1] += mark_thread_hot(subj->thread,
						     node->hot);
		    }
		}
		break;
	    case KillScopeSubthread:
		for (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    if (regexec(node->expr_re, subj->subject,
				0, NULL, 0) == 0) {
			for (art = subj->thread ; art ; ) {
			    if (art->subject == subj) {
				if (node->kill)
				    no[0] += mark_subthread_read(art,
								 False, True);
				else
				    no[1] += mark_subthread_hot(art,
								node->hot);
				art = preorder_skip_subthread(art);
			    } else {
				art = next_in_thread_preorder(art);
			    }
			}
		    }
		}
		break;
	    }
	    break;
	case KillFieldFrom:
	    switch (node->scope) {
	    case KillScopeArticle:
		if (node->kill) {
		    for (art = global.thr.articles ; art ; art = art->next) {
			if (art->from && !art->read &&
			    regexec(node->expr_re, art->from,
				    0, NULL, 0) == 0) {
			    art->read = True;
			    art->killed = True;
			    no[0]++;
			    global.thr.group->no_unread--;
			    art->subject->no_unread--;
			}
		    }
		} else {
		    for (art = global.thr.articles ; art ; art = art->next) {
			if (art->from && art->hot == 0 &&
			    regexec(node->expr_re, art->from,
				    0, NULL, 0) == 0) {
			    art->hot = node->hot;
			    if (!art->read) no[1]++;
			}
		    }
		}
		break;
	    case KillScopeSubject:
		for (art = global.thr.articles ; art ; art = art->next) {
		    if (art->from &&
			regexec(node->expr_re, art->from,
				0, NULL, 0) == 0) {
			if (node->kill)
			    no[0] += mark_subject_read(art->subject,
						       False, True);
			else
			    no[1] += mark_subject_hot(art->subject,
						      node->hot);
		    }
		}
		break;
	    case KillScopeThread:
		for (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    if (subj->prev && subj->prev->thread == subj->thread)
			continue;
		    for (art = subj->thread ; art ;
			 art = next_in_thread_preorder(art)) {
			if (art->from &&
			    regexec(node->expr_re, art->from,
				    0, NULL, 0) == 0) {
			    if (node->kill)
				no[0] += mark_thread_read(subj->thread,
							  False, True);
			    else
				no[1] += mark_thread_hot(subj->thread,
							 node->hot);
			    break;
			}
		    }
		}
		break;
	    case KillScopeSubthread:
		for (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    if (subj->prev && subj->prev->thread == subj->thread)
			continue;
		    for (art = subj->thread ; art ; ) {
			if (art->from &&
			    regexec(node->expr_re, art->from,
				    0, NULL, 0) == 0) {
			    if (node->kill)
				no[0] += mark_subthread_read(art, False, True);
			    else
				no[1] += mark_subthread_hot(art, node->hot);
			    art = preorder_skip_subthread(art);
			} else {
			    art = next_in_thread_preorder(art);
			}
		    }
		}
		break;
	    }
	    break;
	case KillFieldXref:
	    switch (node->scope) {
	    case KillScopeArticle:
		if (node->kill) {
		    for (art = global.thr.articles ; art ; art = art->next) {
			if (art->xref && !art->read &&
			    regexec(node->expr_re, art->xref,
				    0, NULL, 0) == 0) {
			    art->read = True;
			    art->killed = True;
			    global.thr.group->no_unread--;
			    art->subject->no_unread--;
			    no[0]++;
			}
		    }
		} else { /* hot */
		    for (art = global.thr.articles ; art ; art = art->next) {
			if (art->xref && art->hot == 0 &&
			    regexec(node->expr_re, art->xref,
				    0, NULL, 0) == 0) {
			    art->hot = node->hot;
			    if (!art->read) no[1]++;
			}
		    }
		}
		break;
	    case KillScopeSubject:
		for (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    for (art = subj->thread ; art ;
			 art = next_in_thread_preorder(art)) {
			if (art->subject != subj) continue;
			if (art->xref &&
			    regexec(node->expr_re, art->xref,
				    0, NULL, 0) == 0) {
			    if (node->kill)
				no[0] += mark_subject_read(subj, False, True);
			    else
				no[1] += mark_subject_hot(subj, node->hot);
			    break;
			}
		    }
		}
		break;
	    case KillScopeThread:
		for  (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    if (subj->prev && subj->prev->thread == subj->thread)
			continue;
		    for (art = subj->thread ; art ;
			 art = next_in_thread_preorder(art)) {
			if (art->xref &&
			    regexec(node->expr_re, art->xref,
				    0, NULL, 0) == 0) {
			    if (node->kill)
				no[0] += mark_thread_read(subj->thread,
							  False, True);
			    else
				no[1] += mark_thread_hot(subj->thread,
							 node->hot);
			    break;
			}
		    }
		}
		break;
	    case KillScopeSubthread:
		for  (subj = global.thr.subjects ; subj ; subj = subj->next) {
		    if (subj->prev && subj->prev->thread == subj->thread)
			continue;
		    for (art = subj->thread ; art ; ) {
			if (art->xref &&
			    regexec(node->expr_re, art->xref,
				    0, NULL, 0) == 0) {
			    if (node->kill)
				no[0] += mark_subthread_read(art, False, True);
			    else
				no[1] += mark_subthread_hot(art, node->hot);
			    art = preorder_skip_subthread(art);
			} else {
			    art = next_in_thread_preorder(art);
			}
		    }
		}
		break;
	    }
	    break;
	}
    }

    for (subj = global.thr.subjects ; subj ; subj = subj->next) {
	subj->hot = 0; /* just making sure */
	update_subj_hot_value(subj);
    }
}

char *regexp_escape_string(char *src, int anchor)
{
    char	*result, *dest;

    dest = result = XtMalloc(2 * strlen(src) + 3);

    if (anchor)
	*dest++ = '^';
    while (*src != '\0') {
	if (strchr("^.[$()|*+?{\\", *src)) *dest++ = '\\';
	*dest++ = *src++;
    }
    if (anchor)
	*dest++ = '$';
    *dest = '\0';

    return XtRealloc(result, strlen(result) + 1);
}
