/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include "../Widgets/ArtText.h"
#include "../Widgets/ArtTree.h"
#include "../Widgets/Message.h"
#include "../Widgets/Notice.h"
#include "../Widgets/Sash.h"
#include "../Widgets/ScrBar.h"
#include "../Widgets/ScrList.h"
#include "ahead.h"
#include "codes.h"
#include "connect.h"
#include "hash.h"
#include "kedit.h"
#include "kill.h"
#include "newsrc.h"
#include "read.h"
#include "save.h"
#include "search.h"
#include "server.h"
#include "tag.h"
#include "thread.h"
#include "util.h"
#include "widgets.h"
#include "xutil.h"

/* Quit   Done */
void knapp0_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    if (global.busy) return;

    switch (global.mode) {
    case NewsModeDisconnected: /* Quit */
	global_cleanup(False);
	exit(0);
	/* NOTREACHED */
    case NewsModeConnected: /* Quit */
	set_busy(False, True);
	(void)update_newsrc();
	(void)update_kill_file();
	server_write_raw(&global.server, "QUIT\r\n");
	server_close(&global.server);
	exit(0);
	/* NOTREACHED */
    case NewsModeGroup: /* Done */
	set_busy(False, True);

	popdown_save();
	popdown_search();
	clear_tagged_articles();
	clear_history();

	free_read_arts_list(global.thr.group);
	global.thr.group->read_arts = create_read_arts_list();

	thread_ahead_leave_group(global.thr.group);
	global.curr_art = NULL;
	global.curr_subj = NULL;

	if (global.thr.group->subscribed)
	    setNewsModeConnected();
	else
	    setNewsModeAllgroups();

	unset_busy();
	if (global.mode == NewsModeConnected)
	    check_if_rescan_due();
	break;
    case NewsModeAllgroups: /* Done */
	popdown_search();
	setNewsModeConnected();
	check_if_rescan_due();
	break;
    case NewsModeNewgroups: /* Done */
	XtFree((char *)global.new_groups);
	global.new_groups = NULL;
	global.no_new_groups = 0;
	sort_groups();
	setNewsModeConnected();
	check_if_rescan_due();
	break;
    case NewsModeThread:
	break;
    }
}

/* Connect   Disconnect   Subscribe   View Thread  Back */
void knapp1_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    long	i, j;

    if (global.busy) return;

    switch (global.mode) {
    case NewsModeConnected: /* Disconnect */
	set_busy(False, True);

	(void)update_newsrc(); /* should check return value */
	(void)update_kill_file();
	server_write_raw(&global.server, "QUIT\r\n");
	server_close(&global.server);
	global_cleanup(False);
	set_standard_message();

	unset_busy();
	break;
    case NewsModeDisconnected: /* Connect */
	set_standard_message();
	popup_connect_dialogue();
	break;
    case NewsModeThread: /* View Thread */
	set_busy(False, True);
	setNewsModeGroup(False /* not used */);
	unset_busy();
	break;
    case NewsModeGroup: /* Back */
	set_busy(False, True);
	if (global.curr_subj)
	    setNewsModeThread();
	unset_busy();
	break;
    case NewsModeAllgroups: /* Subscribe */
	i = ScrListGetFirstSelected(main_widgets.group_list);

	if (i < 0) {
	    set_message("No selected groups!", True);
	    return;
	}

	set_busy(False, True);
	while (i >= 0) {
	    global.groups[i]->subscribed = True;
	    i = ScrListGetNextSelected(main_widgets.group_list, i);
	}
	sort_groups();
	global.thr.group = NULL;
	setNewsModeAllgroups();
	set_standard_message();
	unset_busy();
	return;
    case NewsModeNewgroups: /* Subscribe */
	i = ScrListGetFirstSelected(main_widgets.group_list);

	if (i < 0) {
	    set_message("No selected groups!", True);
	    return;
	}

	set_busy(False, True);
	while (i >= 0) {
	    for (j = 0 ; j < global.no_groups ; j++) {
		if (global.groups[j]->disp == i) {
		    global.groups[j]->subscribed = True;
		    break;
		}
	    }
	    i = ScrListGetNextSelected(main_widgets.group_list, i);
	}
	setNewsModeNewgroups();
	unset_busy();
	break;
    }
}

/* Unsubscribe   All threads   All groups */
void knapp2_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    long	i, j, n;
    SUBJECT	*subj;

    if (global.busy) return;

    switch (global.mode) {
    case NewsModeConnected: /* All groups */
	set_busy(False, True);
	global.thr.group = NULL;
	setNewsModeAllgroups();
	unset_busy();
	break;
    case NewsModeGroup: /* All threads */
	for (subj = global.thr.subjects ;
	     subj ; subj = subj->next)
	    if (subj->disp < 0)
		break;

	set_curr_art(NULL, True, False);
	set_curr_subj(NULL);
	setNewsModeGroup(subj != NULL);
	set_standard_message();
	break;
    case NewsModeAllgroups: /* Unsubscribe */
	n = ScrListGetFirstSelected(main_widgets.group_list);

	if (n < 0) {
	    set_message("No selected groups!", True);
	    return;
	}

	set_busy(False, True);
	while (n >= 0) {
	    global.groups[n]->subscribed = False;
	    n = ScrListGetNextSelected(main_widgets.group_list, n);
	}
	sort_groups();
	global.thr.group = NULL;
	setNewsModeAllgroups();
	set_standard_message();
	unset_busy();
	break;
    case NewsModeNewgroups: /* Unsubscribe */
	i = ScrListGetFirstSelected(main_widgets.group_list);

	if (i < 0) {
	    set_message("No selected groups!", True);
	    return;
	}

	set_busy(False, True);
	while (i >= 0) {
	    for (j = 0 ; j < global.no_groups ; j++) {
		if (global.groups[j]->disp == i) {
		    global.groups[j]->subscribed = False;
		    break;
		}
	    }
	    i = ScrListGetNextSelected(main_widgets.group_list, i);
	}
	setNewsModeNewgroups();
	unset_busy();
	break;
    case NewsModeDisconnected:
    case NewsModeThread:
	break;
    }
}

/* Kill... */
void knapp5_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    if (global.busy) return;

    popup_kill_editor();
}

/* Update   Catchup */
void knapp6_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    ART_LIST_NODE	*temp;
    int			kill_flag;

    if (global.busy) return;

    switch (global.mode) {
    case NewsModeConnected: /* Update */
    case NewsModeAllgroups:
	set_busy(False, True);

	kill_flag = update_kill_file();
	if (update_newsrc()) {
	    if (kill_flag)
		set_message("Newsrc and kill files updated.", False);
	    else
		set_message("Failed to update kill file! "
			    "  Newsrc file updated.", True);
	} else {
	    if (kill_flag)
		set_message("Failed to update newsrc! "
			    "  Kill file updated.", True);
	    else
		set_message("Failed to update newsrc and kill files!", True);
	}

	unset_busy();
	break;
    case NewsModeGroup: /* Catchup */
	set_busy(False, True);

	popdown_save();
	popdown_search();
	clear_tagged_articles();
	clear_history();

	free_read_arts_list(global.thr.group);
	temp = global.thr.group->read_arts =
	    (ART_LIST_NODE *)XtMalloc(sizeof(ART_LIST_NODE));
	temp->next = NULL;
	temp->first = 1;
	temp->last = global.thr.group->last_art;

	if (GroupObjectProcessXrefs(global.group_object)) {
	    ARTICLE	*art;

	    for (art = global.thr.articles ; art ; art = art->next)
		if (!art->next) process_xref(art);

	}

	thread_ahead_leave_group(global.thr.group);
	global.curr_art = NULL;
	global.curr_subj = NULL;

	setNewsModeConnected();

	unset_busy();
	break;
    case NewsModeDisconnected:
    case NewsModeThread:
    case NewsModeNewgroups:
	break;
    }
}

/* Read/goto group   Next Depth */
void knapp7_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    ARTICLE	*art;
    Boolean	new_thread;

    if (global.busy) return;

    switch (global.mode) {
    case NewsModeConnected: /* Read group */
	set_curr_group();
	/* fall through */
    case NewsModeAllgroups:
	if (!global.thr.group) {
	    set_message("No selected group!", True);
	    return;
	}

	popdown_find_group();
	popdown_search();
	read_group(NULL);
	break;
    case NewsModeGroup:  /* Next Depth */
    case NewsModeThread:
	art = global.curr_art;
	new_thread = False;

	if (!art) {
	    if (!global.curr_subj) {
		set_message("No thread selected!", True);
		return;
	    }
	    art = global.curr_subj->thread;
	    new_thread = True;
	}

	while (art) {
	    if (art->from && !art->read)
		break;
	    art = next_in_thread_preorder(art);
	}

	if (!art) {
	    SUBJECT	*subj;
	    ARTICLE	*stop;

	    new_thread = True;
	    stop = global.curr_subj->thread;
	    subj = global.curr_subj->next;
	    while (subj && subj->thread == stop)
		subj = subj->next;
	    if (!subj)
		subj = global.thr.subjects;

	    while (subj->thread != stop) {
		for (art = subj->thread ; art ;
		     art = next_in_thread_preorder(art))
		    if (art->from && !art->read)
			break;
		if (art)
		    break;

		do {
		    subj = subj->next;
		} while (subj && subj->prev->thread == subj->thread);
		if (!subj)
		    subj = global.thr.subjects;
	    }
	}

	set_curr_art(art, True, True);
	if (art) {
	    if (read_article(art, False, NULL)) {
		if (global.mode == NewsModeThread) {
		    if (new_thread)
			setNewsModeThread();
		} else
		    set_curr_art(art, False, False);
	    }
	} else {
	    if (global.mode == NewsModeThread)
		setNewsModeGroup(False);
	    else
		knapp0_callback(NULL, NULL, NULL); /* Done */
	}

	break;
    case NewsModeDisconnected:
    case NewsModeNewgroups:
	break;
    }
}

/*  Rescan   Previous */
void knapp8_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    ARTICLE	*art;
    Boolean	new_thread;
    char	*buffer;

    if (global.busy) return;

    switch (global.mode) {
    case NewsModeGroup:  /* Previous */
    case NewsModeThread:
	art = history_pop();
	if (global.curr_art && art == global.curr_art)
	    art = history_pop();

	if (!art) {
	    set_message("No previous article!", True);
	    return;
	}

	if (global.curr_art &&
	    global.curr_art->subject->thread == art->subject->thread)
	    new_thread = False;
	else
	    new_thread = True;

	set_curr_art(art, False, True);
	if (read_article(art, False, NULL)) {
	    if (global.mode == NewsModeThread) {
		if (new_thread)
		    setNewsModeThread();
	    } else
		set_curr_art(art, False, False);
	}
	break;
    case NewsModeConnected: /* Rescan */
	remove_rescan_timeout();

	set_busy(True, True);
	set_message("Server contacted, waiting for response...", False);
	buffer = rescan();
	if (!buffer) {
	    (void)reconnect_server(True);
	    unset_busy();
	    return;
	}
	unset_busy();

	if (atoi(buffer) == OK_GROUPS)
	    setNewsModeConnected();
	else {
	    char	message[128];

	    if (strlen(buffer) > 80) buffer[80] = '\0';
	    sprintf(message, "Error!  Failed message from server is: %s",
		    buffer);
	    set_message(message, True);
	}
	break;
    case NewsModeDisconnected:
    case NewsModeAllgroups:
    case NewsModeNewgroups:
	break;
    }
}

/* Abort */
void knapp9_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    if (global.busy && !global.aborted) {
	global.aborted = True;
	if (global.quit_nicely)
	    server_write_raw(&global.server, "\r\nQUIT\r\n");
	server_close(&global.server);
    }
}

/* Save... */
void knapp10_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    if (global.busy) return;

    switch (global.mode) {
    case NewsModeGroup:
    case NewsModeThread:
	popup_save();
	break;
    case NewsModeDisconnected:
    case NewsModeConnected:
    case NewsModeAllgroups:
    case NewsModeNewgroups:
	if (global.bell)
	    XBell(XtDisplay(gw), 0);
	break;
    }
}

/* Search... */
void knapp11_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    if (global.busy) return;

    switch (global.mode) {
    case NewsModeGroup:
    case NewsModeThread:
    case NewsModeAllgroups:
	popup_search();
	break;
    case NewsModeDisconnected:
    case NewsModeConnected:
    case NewsModeNewgroups:
	if (global.bell)
	    XBell(XtDisplay(gw), 0);
	break;
    }
}

void top_vbar_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    ScrBarReport	*report = (ScrBarReport *)call_data;
    long		shown, size;
    Arg			arg;

    switch (global.mode) {
    case NewsModeAllgroups:
    case NewsModeConnected:
    case NewsModeNewgroups:
	ScrListSetFirst(main_widgets.group_list, report->slider_position);
	ScrListGetFirstShownSize(main_widgets.group_list,
				 NULL, &shown, &size);
	if ((report->slider_length != shown ||
	     report->canvas_length != size) &&
	    report->slider_position + shown < size)
	    ScrBarSetLengthsAndPos(main_widgets.top_vbar,
				   size, shown, report->slider_position);
	break;
    case NewsModeGroup:
	ScrListSetFirst(main_widgets.thread_list, report->slider_position);
	ScrListGetFirstShownSize(main_widgets.thread_list,
				 NULL, &shown, &size);
	if ((report->slider_length != shown ||
	     report->canvas_length != size) &&
	    report->slider_position + shown < size)
	    ScrBarSetLengthsAndPos(main_widgets.top_vbar,
				   size, shown, report->slider_position);
	break;
    case NewsModeThread:
	XtSetArg(arg, XtNy, -report->slider_position);
	XtSetValues(main_widgets.arttree, &arg, 1);
	break;
    case NewsModeDisconnected:
	break;
    }
}

void top_hbar_callback(Widget gw, XtPointer client_data, XtPointer call_data)
{
    ScrBarReport	*report = (ScrBarReport *)call_data;

    if (global.mode == NewsModeThread) {
	Arg	arg;

	XtSetArg(arg, XtNx, -report->slider_position);
	XtSetValues(main_widgets.arttree, &arg, 1);
    }
}

void text_scrbar_callback(Widget gw, XtPointer client_data,
			  XtPointer call_data)
{
    ScrBarReport	*report = (ScrBarReport *)call_data;
    long		shown, size;

    ArtTextSetFirst(main_widgets.text, report->slider_position);
    ArtTextGetFirstShownSize(main_widgets.text, NULL, &shown, &size);
    if ((report->slider_length != shown || report->canvas_length != size) &&
	report->slider_position + shown < size)
	ScrBarSetLengthsAndPos(main_widgets.text_scrbar,
			       size, shown, report->slider_position);
}

void top_manager_callback(Widget gw, XtPointer client_data,
			  XtPointer call_data)
{
    long	first, shown, size;
    Position	s_x, s_y;
    Dimension	s_w, s_h, c_w, c_h;
    Arg		args[4];

    switch (global.mode) {
    case NewsModeConnected:
    case NewsModeAllgroups:
    case NewsModeNewgroups:
	ScrListGetFirstShownSize(main_widgets.group_list,
				 &first, &shown, &size);
	ScrBarSetLengthsAndPos(main_widgets.top_vbar, size, shown, first);
	break;
    case NewsModeThread:
	XtSetArg(args[0], XtNx, &s_x);
	XtSetArg(args[1], XtNy, &s_y);
	XtSetArg(args[2], XtNwidth, &c_w);
	XtSetArg(args[3], XtNheight, &c_h);
	XtGetValues(main_widgets.arttree, args, 4);
	XtSetArg(args[0], XtNwidth, &s_w);
	XtSetArg(args[1], XtNheight, &s_h);
	XtGetValues(main_widgets.top_manager, args, 2);
	ScrBarSetLengthsAndPos(main_widgets.top_vbar, c_h, s_h, - s_y);
	ScrBarSetLengthsAndPos(main_widgets.top_hbar, c_w, s_w, - s_x);
	break;
    case NewsModeGroup:
	ScrListGetFirstShownSize(main_widgets.thread_list,
				 &first, &shown, &size);
	ScrBarSetLengthsAndPos(main_widgets.top_vbar, size, shown, first);
	break;
    case NewsModeDisconnected:
	break;
    }
}

void text_resize_callback(Widget gw, XtPointer client_data,
			  XtPointer call_data)
{
    if (global.mode == NewsModeGroup || global.mode == NewsModeThread) {
	ArtTextReport	*report = (ArtTextReport *)call_data;

	ScrBarSetLengthsAndPos(main_widgets.text_scrbar,
			       report->canvas_length,
			       report->slider_length,
			       report->slider_position);
    }
}

void arttree_scroll_callback(Widget gw,
			     XtPointer client_data,
			     XtPointer call_data)
{
    ArtTreeScrollReport	*report = (ArtTreeScrollReport *)call_data;

    if (global.mode == NewsModeThread) {
	ScrBarSetSliderPosition(main_widgets.top_hbar, report->slider_x);
	ScrBarSetSliderPosition(main_widgets.top_vbar, report->slider_y);
    }
}

void thread_list_sel_callback(Widget gw,
			      XtPointer client_data,
			      XtPointer call_data)
{
    long	row = (long)call_data;
    SUBJECT	*loop = global.thr.subjects;

    if (global.busy || global.mode != NewsModeGroup) return;

    global.curr_art = NULL;
    if (global.curr_subj && global.curr_subj->disp == row) return;

    while (loop) {
	if (loop->disp == row) break;
	loop = loop->next;
    }
    global.curr_subj = loop;
}

void thread_list_callback(Widget gw,
			  XtPointer client_data,
			  XtPointer call_data)
{
    long	row = (long)call_data;
    SUBJECT	*subj;
    ARTICLE	*art = NULL;

    if (global.busy || global.mode != NewsModeGroup) return;

    if (!global.curr_subj || global.curr_subj->disp != row) {
	for (subj = global.curr_subj ; subj ; subj = subj->next) {
	    if (subj->disp == row) break;
	}
	if (!subj) return;
	global.curr_subj = subj;
    } else subj = global.curr_subj;

    for (art = subj->thread ; art ; art = next_in_thread_preorder(art)) {
	if (art->subject == subj && art->from && !art->read) break;
    }

    global.curr_art = art;
    if (art)
	history_push(art);
    (void)read_article(art, False, NULL);

    if (!art) {
	if (global.bell)
	    XBell(XtDisplay(gw), 0);
	set_message("No unread article with that subject!", True);
    }
}

void arttree_sel_callback(Widget gw,
			  XtPointer client_data,
			  XtPointer call_data)
{
    ARTICLE	*art = (ARTICLE *)call_data;

    if (global.busy) return;

    if (global.mode == NewsModeThread) {
	set_curr_art(art, True, False);
	(void)read_article(art, False, NULL);
    }	
}

void group_list_callback(Widget gw,
			 XtPointer client_data,
			 XtPointer call_data)
{
    long	row = (long)call_data;

    if (global.busy) return;

    global.thr.group = NULL;

    if (global.mode == NewsModeAllgroups) {
	popdown_search();
	global.thr.group = global.groups[row];
    } else if (global.mode == NewsModeConnected) {
	GROUP_NODE	**loop = global.groups;
	long		n = global.no_groups;

	while (n > 0 && (*loop)->subscribed) {
	    if ((*loop)->disp == row) {
		global.thr.group = *loop;
		break;
	    }
	    loop++;
	    n--;
	}

	if (!global.thr.group) return;
    } else {
	return;
    }

    read_group(NULL);
}

void group_list_sel_callback(Widget w,
			     XtPointer client_data,
			     XtPointer call_data)
{
    long	row;

    if (global.busy) return;

    switch (global.mode) {
    case NewsModeDisconnected:
    case NewsModeNewgroups:
    case NewsModeGroup:
    case NewsModeThread:
	break;
    case NewsModeConnected:
	row = (long)call_data;
	if (row >= 0) {
	    long	i;

	    for (i = 0 ; i < global.no_groups ; i++) {
		if (!global.groups[i]->subscribed)
		    break;
		if (global.groups[i]->disp == row) {
		    global.thr.group = global.groups[i];
		    return;
		}
	    }
	    global.thr.group = NULL;
	}
	break;
    case NewsModeAllgroups:
	row = (long)call_data;
	if (row >= 0 && row < global.no_groups)
	    global.thr.group = global.groups[row];
	else
	    global.thr.group = NULL;
	break;
    }
}

void group_list_dnd_callback(Widget w,
			     XtPointer client_data,
			     XtPointer call_data)
{
    long	*index = (long *)call_data;
    GROUP_NODE	*temp;
    long	i;

    index[2] = False;
    switch (global.mode) {
    case NewsModeAllgroups:
	if (global.groups[index[0]]->subscribed &&
	    global.groups[index[1]]->subscribed) {
	    index[2] = True;
	    temp = global.groups[index[0]];
	    if (index[0] < index[1])
		for (i = index[0] ; i < index[1] ; i++)
		    global.groups[i] = global.groups[i + 1];
	    else
		for (i = index[0] ; i > index[1] ; i--)
		    global.groups[i] = global.groups[i - 1];
	    global.groups[index[1]] = temp;
	    break;
	}
	/* fall through */
    case NewsModeDisconnected:
    case NewsModeGroup:
    case NewsModeThread:
    case NewsModeConnected:
    case NewsModeNewgroups:
	if (global.bell)
	    XBell(XtDisplay(w), 0);
	break;
    }
}

void delete_window_callback(Widget w,
			    XtPointer client_data,
			    XtPointer call_data)
{
    if (global.quit_nicely || !global.busy)
	server_write_raw(&global.server, "\r\nQUIT\r\n");
    global_cleanup(False);
    exit(0);
}

void sash_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
    SashCallDataRec	*data = (SashCallDataRec *)call_data;

    if (data && data->event->type == MotionNotify) {
	long	y = data->event->xmotion.y;

	if (y != 0) {
	    Dimension	height, width;
	    Arg		arg[2];

	    XtSetArg(arg[0], XtNheight, &height);
	    XtSetArg(arg[1], XtNwidth, &width);
	    XtGetValues(main_widgets.top_layout, arg, 2);
	    y += height;
	    if (y <= 0)
		y = 1;
	    else if (y > 32767)
		y = 32767;
	    (void)XtMakeResizeRequest(main_widgets.top_layout,
				      width, y, NULL, NULL);
	}
    }
}
