/* This software is Copyright 1995 by Karl-Johan Johnsson
 *
 * Permission is hereby granted to copy, reproduce, redistribute or otherwise
 * use this software as long as: there is no monetary profit gained
 * specifically from the use or reproduction of this software, it is not
 * sold, rented, traded or otherwise marketed, and this copyright notice is
 * included prominently in any copy made. 
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. ANY USE OF THIS
 * SOFTWARE IS AT THE USER'S OWN RISK.
 */
#include "global.h"
#include "ahead.h"
#include "codes.h"
#include "connect.h"
#include "hash.h"
#include "kedit.h"
#include "kill.h"
#include "newsrc.h"
#include "save.h"
#include "search.h"
#include "server.h"
#include "tag.h"
#include "util.h"
#include "widgets.h"
#include "xutil.h"
#include "../Widgets/ArtTree.h"
#include "../Widgets/Message.h"
#include "../Widgets/Notice.h"
#include "../Widgets/ScrBar.h"
#include "../Widgets/ScrList.h"
#include "../Widgets/Shadow.h"

void set_curr_art(ARTICLE *art, int push, int center)
{
    if (art && push)
	history_push(art);

    if (global.mode == NewsModeThread) {
	if (global.curr_art && global.curr_art != art)
	    ArtTreeNodeSetSelected(main_widgets.arttree,
				   (ART_TREE_NODE *)global.curr_art, False);

	if (art) {
	    ArtTreeNodeSetSelected(main_widgets.arttree,
				   (ART_TREE_NODE *)art, True);
	    if (center)
		ArtTreeNodeMakeVisible(main_widgets.arttree,
				       (ART_TREE_NODE*)art);
	}
    }

    global.curr_art = art;
    if (art)
	set_curr_subj(art->subject);
}

void set_curr_subj(SUBJECT *subj)
{
    global.curr_subj = subj;
    if (subj && subj->disp >= 0) {
	long	first, shown, size;
	long	new = subj->disp;

	ScrListSetSelected(main_widgets.thread_list, new, True);
	ScrListGetFirstShownSize(main_widgets.thread_list,
				 &first, &shown, &size);
	if (new < first || first + shown <= new) {
	    if (new + shown >= size) {
		new = size - shown;
		if (new < 0)
		    new = 0;
	    }
	    ScrListSetFirst(main_widgets.thread_list, new);
	    if (global.mode == NewsModeGroup)
		ScrBarSetSliderPosition(main_widgets.top_vbar, new);
	}
    }
}

void set_tree_stuff(ARTICLE *thr)
{
    while (thr) {
	if (thr->from && !thr->read) {
	    ArtTreeNodeSetInner(main_widgets.arttree,
				(ART_TREE_NODE *)thr, True);
	    if (thr->hot > 0)
		ArtTreeNodeSetPixmap(main_widgets.arttree,
				     (ART_TREE_NODE *)thr,
				     global.hot_pixmaps[thr->hot]);
	}
	if (A_PARENT(thr) && A_PARENT(thr)->subject != thr->subject) {
	    ArtTreeNodeSetDashed(main_widgets.arttree,
				 (ART_TREE_NODE*)thr, True);
	}

	if (A_CHILD1(thr))
	    set_tree_stuff(A_CHILD1(thr));
	thr = A_SIBLING(thr);
    }
}

void set_curr_group()
{
    long	i, n;

    switch (global.mode) {
    case NewsModeConnected:
    case NewsModeAllgroups:
    case NewsModeNewgroups:
	n = ScrListGetFirstSelected(main_widgets.group_list);
	if (!global.thr.group || global.thr.group->disp != n) {
	    global.thr.group = NULL;

	    if (n < 0)
		break;

	    for (i = 0 ; i < global.no_groups ; i++) {
		if (!global.groups[i]->subscribed)
		    break;
		else if (global.groups[i]->disp == n) {
		    global.thr.group = global.groups[i];
		    break;
		}
	    }
	}
	break;
    case NewsModeGroup:
    case NewsModeThread:
    case NewsModeDisconnected:
	break;
    }
}

void global_cleanup(int notify)
{
    long	i;

    if (notify)
	set_message("Error: Connection to server broken!", True);

    server_close(&global.server);

    if (global.mode == NewsModeGroup ||
	global.mode == NewsModeThread) {
	free_read_arts_list(global.thr.group);
	global.thr.group->read_arts = create_read_arts_list();
	clear_thread_context(&global.thr);
    }

    thread_ahead_shutdown();
    popdown_find_group();
    popdown_kill_editor();
    popdown_save();
    popdown_search();
    clear_tagged_articles();
    clear_history();

    if (global.mode != NewsModeDisconnected) {
	(void)update_newsrc();
	(void)update_kill_file();
    }

    free_kill_file();

    for (i = 0 ; i < global.no_groups ; i++) {
	XtFree(global.groups[i]->name);
	XtFree(global.groups[i]->description);
	free_read_arts_list(global.groups[i]);
	XtFree((char *)global.groups[i]);
    }
    XtFree((char *)global.groups);
    global.groups = NULL;
    global.no_groups = 0;
    global.max_groups = 0;
    global.thr.group = NULL;
    XtFree((char *)global.new_groups);
    global.new_groups = NULL;
    global.no_new_groups = 0;
    
    global.curr_art = NULL;
    global.curr_subj = NULL;

    setNewsModeDisconnected();
}

void change_interruptible(int interruptible)
{
    Arg	arg;

    if (!global.busy) {
	fprintf(stderr, "change_iunterruptible, but not busy!\n");
	return;
    }

    arg.name = XtNsensitive;
    arg.value = (XtArgVal)interruptible;
    XtSetValues(main_widgets.knapp[9], &arg, 1);
    ShadowRedrawWidget(main_widgets.knapp[9]);
    XFlush(XtDisplay(main_widgets.shell));
}

void set_busy(int interruptible, int quit_nicely)
{
    Display	*disp = XtDisplay(main_widgets.shell);

    if (global.busy) {
	fprintf(stderr, "warning multiple busy\n");
	return;
    }

    global.busy = True;
    global.quit_nicely = quit_nicely;
    global.aborted = False;

    XDefineCursor(disp, XtWindow(main_widgets.shell), global.busy_cursor);
    if (main_widgets.kill_editor_shell)
	XDefineCursor(disp,
	    XtWindow(main_widgets.kill_editor_shell),
	    global.busy_cursor);
    if (main_widgets.save_shell)
	XDefineCursor(disp,
		      XtWindow(main_widgets.save_shell),
		      global.busy_cursor);
    if (main_widgets.search_shell)
	XDefineCursor(disp,
		      XtWindow(main_widgets.search_shell),
		      global.busy_cursor);
    if (main_widgets.second_shell)
	XDefineCursor(disp,
		      XtWindow(main_widgets.second_shell),
		      global.busy_cursor);

    if (interruptible) {
	XtSetSensitive(main_widgets.knapp[9], interruptible);
	ShadowRedrawWidget(main_widgets.knapp[9]);
    }

    XFlush(XtDisplay(main_widgets.shell));
    
    XtAddGrab(main_widgets.knapp[9], True, False);
}

void unset_busy()
{
    Display	*disp = XtDisplay(main_widgets.shell);

    if (!global.busy) {
	fprintf(stderr, "warning: multiple unset_busy\n");
	return;
    }

    global.busy = False;

    XDefineCursor(disp, XtWindow(main_widgets.shell), global.cursor);
    if (main_widgets.kill_editor_shell)
	XDefineCursor(disp,
		      XtWindow(main_widgets.kill_editor_shell),
		      global.cursor);
    if (main_widgets.save_shell)
	XDefineCursor(disp,
		      XtWindow(main_widgets.save_shell),
		      global.cursor);
    if (main_widgets.search_shell)
	XDefineCursor(disp,
		      XtWindow(main_widgets.search_shell),
		      global.cursor);
    if (main_widgets.second_shell)
	XDefineCursor(disp,
		      XtWindow(main_widgets.second_shell),
		      global.cursor);

    XtSetSensitive(main_widgets.knapp[9], False);
    ShadowRedrawWidget(main_widgets.knapp[9]);

    XFlush(disp);

    XtRemoveGrab(main_widgets.knapp[9]);
}

void set_standard_message()
{
    char	message[512];
    long	n;

    switch (global.mode) {
    case NewsModeDisconnected:
	set_message("Disconnected.", False);
	break;
    case NewsModeConnected:
	if (global.groups && global.groups[0] &&
	    global.groups[0]->subscribed)
	    set_message("Subscribed groups "
			"with unread articles displayed.",
			False);
	else
	    set_message("No subscribed groups.", False);
	break;
    case NewsModeGroup:
	ScrListGetFirstShownSize(main_widgets.thread_list, NULL, NULL, &n);
	sprintf(message, "%s with %ld unread articles, %ld subjects.",
		global.thr.group->name, global.thr.group->no_unread, n);
	set_message(message, False);
	break;
    case NewsModeThread:
	if (!global.curr_subj)
	    return;
	sprintf(message, "%ld unread articles with Subject: ",
		global.curr_subj->no_unread);
	strncat(message, global.curr_subj->subject, 128);
	set_message(message, False);
	break;
    case NewsModeAllgroups:
	set_message("All groups displayed.", False);
	break;
    case NewsModeNewgroups:
	set_message("New groups displayed.", False);
	break;
    }
}

void stderr_popup(char *stderr_buf, long timeout)
{
    if (timeout < 0)
	timeout = global.stderr_timeout;

    if (stderr_buf && stderr_buf[0] != '\0' && timeout >= 0)
	NoticeCreateAndPopup("stderr", main_widgets.shell,
			     stderr_buf, "Ok", NULL, NULL,
			     timeout, NULL, NULL, XtGrabNone);
}

/*****************************************************************************/

static XtIntervalId	rescan_timeout_id = 0;
static Boolean		rescan_due = False;

static void rescan_timeout_callback(XtPointer client_data, XtIntervalId *id)
{
    char	*buffer;

    rescan_timeout_id = 0;
    rescan_due = False;
    if (global.mode == NewsModeDisconnected ||
	NntpObjectRescanTimeout(global.nntp_object) <= 0)
	return;

    if (global.mode != NewsModeConnected || global.busy) {
	rescan_due = True;
	return; /* no op */
    }

    set_busy(True, True);
    set_message("Automatic rescan in progress...", False);
    buffer = rescan();
    if (!buffer) {
	(void)reconnect_server(True);
	unset_busy();
	return;
    }
    unset_busy();

    if (atoi(buffer) == OK_GROUPS)
	setNewsModeConnected();
    else {
	char	message[128];

	if (strlen(buffer) > 80) buffer[80] = '\0';
	sprintf(message, "Error!  Failed message from server is: %s",
		buffer);
	set_message(message, True);
    }
}

void add_rescan_timeout()
{
    int	timeout = NntpObjectRescanTimeout(global.nntp_object);

    rescan_due = False;

    if (rescan_timeout_id != 0)
	XtRemoveTimeOut(rescan_timeout_id);

    if (timeout > 0)
	rescan_timeout_id =
	    XtAppAddTimeOut(global.app_cont, timeout * 60000L,
			    rescan_timeout_callback, NULL);
    else
	rescan_timeout_id = 0;	
}

void remove_rescan_timeout()
{
    if (rescan_timeout_id != 0)
	XtRemoveTimeOut(rescan_timeout_id);
    rescan_timeout_id = 0;
    rescan_due = False;
}

void check_if_rescan_due()
{
    int	timeout = NntpObjectRescanTimeout(global.nntp_object);

    if (timeout <= 0)
	return;

    if (rescan_timeout_id == 0) {
	if (rescan_due)
	    rescan_timeout_id = XtAppAddTimeOut(global.app_cont, 2000,
						rescan_timeout_callback, NULL);
	else
	    add_rescan_timeout();
    }
}
